package bar.stundenplan;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.Toolkit;
import java.util.Random;
import java.util.Set;
import java.util.prefs.Preferences;
import javax.swing.JComboBox;
import javax.swing.JFrame;

/**
 * @author Benjamin A. Bartsch
 */
public class Config {

    private static final Preferences prefs = Preferences.userRoot().node(Config.class.getName());

    static Font FONT_HEADER = new Font(Font.SANS_SERIF, Font.BOLD, 18);
    static Font FONT_NORMAL_SANS;
    static Font FONT_NORMAL_MONO;
    static Font FONT_NORMAL_HEADER;

    static final int STUNDEN_TYP_N = 1;
    static final int STUNDEN_TYP_D = 2;
    static final int STUNDEN_TYP_X = 3;
    static final int SOLVER_SAT = 1;
    static final int SOLVER_UNSAT = -1;
    static final int SOLVER_UNKNOWN = 0;
    static final String PLAN_P_P = "++";
    static final String PLAN_P_M = "+-";
    static final String PLAN_M_P = "-+";
    static final String PLAN_M_M = "--";
    static final int STUNDEN_LAGE_EGAL = 0;
    static final int STUNDEN_LAGE_AM_ANFANG = 1;
    static final int STUNDEN_LAGE_NICHT_AM_ANFANG = 2;
    static final int STUNDEN_LAGE_AM_ENDE = 3;
    static final int STUNDEN_LAGE_NICHT_AM_ENDE = 4;
    static final int STUNDEN_LAGE_AM_RAND = 5;
    static final int STUNDEN_LAGE_NICHT_AM_RAND = 6;
    static final int KURSE_SORTIEREN_NACH_KL_FA_LE = 0;
    static final int KURSE_SORTIEREN_NACH_KL_LE_FA = 1;
    static final int KURSE_SORTIEREN_NACH_LE_KL_FA = 2;
    static final int KURSE_SORTIEREN_NACH_LE_FA_KL = 3;
    static final int KURSE_SORTIEREN_NACH_FA_KL_LE = 4;
    static final int KURSE_SORTIEREN_NACH_FA_LE_KL = 5;

    static {
        Dimension screenSize = Toolkit.getDefaultToolkit().getScreenSize();
        double width = screenSize.getWidth();
        double height = screenSize.getHeight();
        FONT_NORMAL_SANS = new Font(Font.SANS_SERIF, Font.PLAIN, (int) (Math.min(width, height) / 70));
        FONT_NORMAL_MONO = new Font(Font.MONOSPACED, Font.PLAIN, (int) (Math.min(width, height) / 70));
        FONT_NORMAL_HEADER = new Font(Font.DIALOG, Font.BOLD, (int) (Math.min(width, height) / 70));
    }

    public static void main(String[] args) {
        Random rnd = new Random();
        System.out.println("" + rnd.nextLong());
    }

    public static String get_solver_crypto_pfad() {
        return prefs.get("cryptominisat_pfad", "Bitte den Pfad auswählen...");
    }

    static String get_solver_speicher_pfad() {
        return prefs.get("speicher_pfad", "Bitte den Pfad auswählen...");
    }

    public static void set_solver_crypto_pfad(String pfad) {
        prefs.put("cryptominisat_pfad", pfad);
    }

    static void set_solver_speicher_pfad(String pfad) {
        prefs.put("speicher_pfad", pfad);
    }

    static int get_first_empty_id(Set<Integer> keySet) throws Exception {
        for (int id = 1; id <= 10000; id++) {
            if (keySet.contains(id) == false) {
                return id;
            }
        }
        throw new Exception("get_first_empty_id konnte keine ID finden.");
    }

    static void add_items(JComboBox<String> cb, boolean select_empty, int min, int max) {
        cb.removeAllItems();
        cb.addItem("");
        for (int i = min; i <= max; i++) {
            cb.addItem("" + i);
        }
        if (select_empty) {
            cb.setSelectedIndex(0);
        }
    }

    static void add_items_plan_inner(JComboBox<String> cb, boolean select_empty) {
        cb.removeAllItems();
        cb.addItem("");
        cb.addItem("++");
        cb.addItem("--");
        cb.addItem("+-");
        cb.addItem("-+");
        if (select_empty) {
            cb.setSelectedIndex(0);
        }
    }

    static void setFileChooserFont(Component[] comp, Font font) {
        for (int x = 0; x < comp.length; x++) {
            if (comp[x] instanceof Container) {
                setFileChooserFont(((Container) comp[x]).getComponents(), font);
            }
            try {
                comp[x].setFont(font);
            } catch (Exception e) {
            }
        }
    }

    static void screen_center(JFrame frame, double percent) {
        Dimension dim1 = Toolkit.getDefaultToolkit().getScreenSize();
        Dimension dim2 = new Dimension((int) (dim1.width * percent), (int) (dim1.height * percent));
        frame.setSize(dim2);
        int x = (int) ((dim1.getWidth() - frame.getWidth()) / 2);
        int y = (int) ((dim1.getHeight() - frame.getHeight()) / 2);
        frame.setLocation(x, y);
    }

}
