package bar.stundenplan;

import java.util.LinkedList;

/**
 * @author Benjamin A. Bartsch
 */
public class D_Klasse_SAT {

    final int MAX_W = 2;
    final int MAX_T;
    final int MAX_S;

    final int id;
    final String kuerzel;
    final String plan;

    final int[][][] planWTS; // Woche, Tag, Stunde
    final int[][] planWT_am_tag_unterricht; // Unterricht ist am Tag ...
    final int[][][] planWTS_hier_oder_davor_unterricht; // Woche, Tag, Stunde
    final int[][][] planWTS_hier_oder_danach_unterricht; // Woche, Tag, Stunde

    final LinkedList<D_Kurs_SAT> myKu = new LinkedList<>();

    public D_Klasse_SAT(int MAX_T, int MAX_S, D_Klasse kl) {
        this.MAX_T = MAX_T;
        this.MAX_S = MAX_S;

        this.id = kl.id;
        this.kuerzel = kl.kuerzel;
        this.plan = kl.plan;

        this.planWTS = new int[MAX_W][MAX_T][MAX_S];
        this.planWT_am_tag_unterricht = new int[MAX_W][MAX_T];
        this.planWTS_hier_oder_davor_unterricht = new int[MAX_W][MAX_T][MAX_S];
        this.planWTS_hier_oder_danach_unterricht = new int[MAX_W][MAX_T][MAX_S];
    }

    void cnf_pre(SatWrapper sw) {
        // this
        sw.c_fill_var(planWTS);
        sw.c_fill_var(planWT_am_tag_unterricht);
        sw.c_fill_var(planWTS_hier_oder_davor_unterricht);
        sw.c_fill_var(planWTS_hier_oder_danach_unterricht);
    }

    void cnf_post(SatWrapper sw) throws Exception {
        // this: planWT_am_tag_unterricht
        for (int w = 0; w < MAX_W; w++) {
            for (int t = 0; t < MAX_T; t++) {
                LinkedList<Integer> list = new LinkedList<>();
                for (int s = 0; s < MAX_S; s++) {
                    list.addLast(planWTS[w][t][s]);
                }
                int root = sw.c_OR_tree(list);
                sw.c_equal(root, planWT_am_tag_unterricht[w][t]);
            }
        }
        // this: planWTS (maximal 1 Kurs oder eine Kopplung)
        for (int w = 0; w < MAX_W; w++) {
            for (int t = 0; t < MAX_T; t++) {
                for (int s = 0; s < MAX_S; s++) {
                    // planWTS
                    LinkedList<D_Kopplung_SAT> list_ko = new LinkedList<>();
                    LinkedList<Integer> listVAR = new LinkedList<>();
                    for (D_Kurs_SAT ku_sat : myKu) {
                        if (ku_sat.myKo == null) {
                            // Kurs hinzufügen 
                            listVAR.addLast(ku_sat.planWTS[w][t][s]);
                        } else {
                            // Kopplung hinzufügen 
                            if (!list_ko.contains(ku_sat.myKo)) {
                                list_ko.addLast(ku_sat.myKo);
                                listVAR.addLast(ku_sat.myKo.planWTS_klasse[w][t][s]);
                            }
                        }
                    }
                    int root = sw.c_at_most_one_tree(listVAR);
                    sw.c_equal(root, planWTS[w][t][s]);
                    // planWTS_hier_oder_davor_unterricht
                    if (s == 0) {
                        sw.c_equal(planWTS_hier_oder_davor_unterricht[w][t][s], planWTS[w][t][s]);
                    } else {
                        int x = planWTS_hier_oder_davor_unterricht[w][t][s - 1];
                        int y = planWTS[w][t][s];
                        sw.c_equal(planWTS_hier_oder_davor_unterricht[w][t][s], sw.c_new_var_OR(x, y));
                    }
                    // planWTS_hier_oder_danach_unterricht
                    if (s == MAX_S - 1) {
                        sw.c_equal(planWTS_hier_oder_danach_unterricht[w][t][s], planWTS[w][t][s]);
                    } else {
                        int x = planWTS_hier_oder_danach_unterricht[w][t][s + 1];
                        int y = planWTS[w][t][s];
                        sw.c_equal(planWTS_hier_oder_danach_unterricht[w][t][s], sw.c_new_var_OR(x, y));
                    }
                }
            }
        }
        // plan
        sw.c_plan(plan, planWTS, MAX_T, MAX_S, false, false, false, false, false, false, false);
        
        // lage eines stundenelementes eines kurses
        for (D_Kurs_SAT ku_sat : myKu) {
            for (D_Stunde_SAT st_sat : ku_sat.mySt.rows()) {
                if (st_sat.lage_kl != Config.STUNDEN_LAGE_EGAL) {
                    st_sat.cnf_post_plan_lage(sw, st_sat.lage_kl, planWTS_hier_oder_davor_unterricht, planWTS_hier_oder_danach_unterricht);
                }
            }
        }
        
    }

    @Override
    public int hashCode() {
        return id;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final D_Klasse_SAT other = (D_Klasse_SAT) obj;
        if (this.id != other.id) {
            return false;
        }
        return true;
    }
    
    
}
