package bar.stundenplan;

import java.util.LinkedList;

/**
 * @author Benjamin A. Bartsch
 */
public class D_Kopplung_SAT {

    final int MAX_W = 2;
    final int MAX_T;
    final int MAX_S;

    final int id;
    final String kuerzel;
    final String anzeigename;
    final String plan;

    final int[][][] planWTS_implication; // Woche, Tag, Stunde
    final int[][][] planWTS_klasse; // Woche, Tag, Stunde
    final int[][] planWT_ist_am; // Unterricht ist am Tag ...

    final LinkedList<D_Kurs_SAT> myKu = new LinkedList<>();
    final RowMap<D_Stunde_SAT> mySt = new RowMap<>();

    public D_Kopplung_SAT(int MAX_T, int MAX_S, D_Kopplung ko) throws Exception {
        this.MAX_T = MAX_T;
        this.MAX_S = MAX_S;

        this.id = ko.id;
        this.kuerzel = ko.kuerzel;
        this.anzeigename = ko.anzeigename;
        this.plan = ko.plan;

        this.planWTS_implication = new int[MAX_W][MAX_T][MAX_S];
        this.planWTS_klasse = new int[MAX_W][MAX_T][MAX_S];
        this.planWT_ist_am = new int[MAX_W][MAX_T];

        // Stunden kopieren
        for (int row = 0; row < ko.st_map.size(); row++) {
            D_Stunde_Ko st = ko.st_map.get_obj_with_row(row);

            if (st.typ == Config.STUNDEN_TYP_N) {
                mySt.put(st.id, new D_Stunde_SAT_n(MAX_T, MAX_S, st.id, st.wochenstunden, st.plan, st.lage_kl, st.lage_le));
                continue;
            }

            if (st.typ == Config.STUNDEN_TYP_D) {
                mySt.put(st.id, new D_Stunde_SAT_d(MAX_T, MAX_S, st.id, st.wochenstunden, st.plan, st.lage_kl, st.lage_le));
                continue;
            }

            throw new Exception("D_Kopplung_SAT: Stundentyp " + st.typ + " (x) ist unbekannt!");
        }
    }

    @Override
    public int hashCode() {
        return id;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final D_Kopplung_SAT other = (D_Kopplung_SAT) obj;
        if (this.id != other.id) {
            return false;
        }
        return true;
    }

    void println_debug(SatWrapper sw) {
        System.out.println("");
        System.out.println("Kopplung " + kuerzel);
        System.out.print("h: ");
        for (int t = 0; t < MAX_T; t++) {
            String z0 = sw.is_true(planWT_ist_am[0][t]) ? "1" : "0";
            String z1 = sw.is_true(planWT_ist_am[1][t]) ? "1" : "0";
            String zelle;
            if (z0.equals(z1)) {
                zelle = "  " + z0 + "  ";
            } else {
                zelle = " " + z0 + "/" + z1 + " ";
            }
            System.out.print(zelle);
        }
        System.out.println("");
        System.out.println("");

        for (int s = 0; s < MAX_S; s++) {
            System.out.print((s + 1) + ": ");
            for (int t = 0; t < MAX_T; t++) {
                String z0 = sw.is_true(planWTS_implication[0][t][s]) ? "1" : "0";
                String z1 = sw.is_true(planWTS_implication[1][t][s]) ? "1" : "0";
                String zelle;
                if (z0.equals(z1)) {
                    zelle = "  " + z0 + "  ";
                } else {
                    zelle = " " + z0 + "/" + z1 + " ";
                }
                System.out.print(zelle);
            }
            System.out.println("");
        }

        System.out.println("");
        System.out.println("");

        for (int s = 0; s < MAX_S; s++) {
            System.out.print((s + 1) + ": ");
            for (int t = 0; t < MAX_T; t++) {
                String z0 = sw.is_true(planWTS_klasse[0][t][s]) ? "1" : "0";
                String z1 = sw.is_true(planWTS_klasse[1][t][s]) ? "1" : "0";
                String zelle;
                if (z0.equals(z1)) {
                    zelle = "  " + z0 + "  ";
                } else {
                    zelle = " " + z0 + "/" + z1 + " ";
                }
                System.out.print(zelle);
            }
            System.out.println("");
        }

    }

    void cnf_pre(SatWrapper sw) {
        // this
        sw.c_fill_var(planWTS_implication);
        sw.c_fill_var(planWTS_klasse);
        sw.c_fill_var(planWT_ist_am);

        // Stunden 
        for (int row = 0; row < mySt.size(); row++) {
            D_Stunde_SAT st = mySt.get_obj_with_row(row);
            st.cnf_pre(sw);
        }
    }

    void cnf_post(SatWrapper sw, D_Schule_SAT sc) throws Exception {
        // this: Jeder Stundenblock hat seinen eigenen Tag
        for (int w = 0; w < MAX_W; w++) {
            for (int t = 0; t < MAX_T; t++) {
                LinkedList<Integer> list = new LinkedList<>();
                for (int row = 0; row < mySt.size(); row++) {
                    D_Stunde_SAT st = mySt.get_obj_with_row(row);
                    list.addLast(st.planWT_ist_am[w][t]);
                }
                int root = sw.c_at_most_one_tree(list);
                sw.c_equal(root, planWT_ist_am[w][t]);
            }
        }

        // planWTS_implication
        for (int w = 0; w < MAX_W; w++) {
            for (int t = 0; t < MAX_T; t++) {
                for (int s = 0; s < MAX_S; s++) {
                    LinkedList<Integer> list = new LinkedList<>();
                    for (int row = 0; row < mySt.size(); row++) {
                        D_Stunde_SAT st = mySt.get_obj_with_row(row);
                        list.addLast(st.planWTS[w][t][s]);
                    }
                    int root = sw.c_at_most_one_tree(list);
                    sw.c_equal(root, planWTS_implication[w][t][s]);
                }
            }
        }

        // ku --> planWTS_implication
        // planWTS_klasse
        for (int w = 0; w < MAX_W; w++) {
            for (int t = 0; t < MAX_T; t++) {
                for (int s = 0; s < MAX_S; s++) {
                    // Implikationen von Kurs
                    for (D_Kurs_SAT ku_sat : myKu) {
                        int x = ku_sat.planWTS[w][t][s];
                        int y = planWTS_implication[w][t][s];
                        sw.c_implication(x, y);
                    }
                    // planWTS_klasse
                    LinkedList<Integer> list_ku = new LinkedList<>();
                    for (D_Kurs_SAT ku_sat : myKu) {
                        list_ku.addLast(ku_sat.planWTS[w][t][s]);
                    }
                    int root = sw.c_OR_tree(list_ku);
                    sw.c_equal(root, planWTS_klasse[w][t][s]);
                }
            }
        }

        // Stunden 
        for (int row = 0; row < mySt.size(); row++) {
            D_Stunde_SAT st = mySt.get_obj_with_row(row);
            st.cnf_post(sw, sc);
        }

        // plan
        sw.c_plan(plan, planWTS_klasse, MAX_T, MAX_S, false, false, false, false, false, false, false);
    }

}
