package bar.stundenplan;

import java.util.HashMap;
import java.util.LinkedList;

/**
 * @author Benjamin A. Bartsch
 */
public class D_Kurs_SAT {

    final int MAX_W = 2;
    final int MAX_T;
    final int MAX_S;
    final int id;
    final String ergaenzung;
    final String plan;
    final int vertretungsreserven;
    final int nicht_aufeinanderfolgend;

    D_Kopplung_SAT myKo = null;
    final int[][][] planWTS; // Woche, Tag, Stunde
    final int[][] planWT_ist_am; // Unterricht ist am Tag ...

    final RowMap<D_Stunde_SAT> mySt = new RowMap<>();
    final HashMap<D_Raum_SAT, Integer> planR = new HashMap<>();
    final LinkedList<D_Kurs_SAT> nicht_gleicher_tag_wie = new LinkedList<>();

    public D_Kurs_SAT(int MAX_TAGE, int MAX_STUNDEN, D_Kurs ku) throws Exception {
        this.MAX_T = MAX_TAGE;
        this.MAX_S = MAX_STUNDEN;

        this.id = ku.id;
        this.ergaenzung = ku.ergaenzung;
        this.plan = ku.plan;
        this.vertretungsreserven = ku.vertretungsreserven;
        this.nicht_aufeinanderfolgend = ku.nicht_aufeinanderfolgend;

        this.planWTS = new int[MAX_W][MAX_TAGE][MAX_STUNDEN];
        this.planWT_ist_am = new int[MAX_W][MAX_TAGE];
        // Stunden kopieren
        for (int row = 0; row < ku.st_map.size(); row++) {
            D_Stunde st = ku.st_map.get_obj_with_row(row);

            if (st.typ == Config.STUNDEN_TYP_N) {
                mySt.put(st.id, new D_Stunde_SAT_n(MAX_TAGE, MAX_STUNDEN, st.id, st.wochenstunden, st.plan, st.lage_kl, st.lage_le));
                continue;
            }

            if (st.typ == Config.STUNDEN_TYP_D) {
                mySt.put(st.id, new D_Stunde_SAT_d(MAX_TAGE, MAX_STUNDEN, st.id, st.wochenstunden, st.plan, st.lage_kl, st.lage_le));
                continue;
            }

            throw new Exception("D_Kurs_SAT: Stundentyp " + st.typ + " (x) ist unbekannt!");
        }

    }

    void cnf_pre(SatWrapper sw) {
        // this
        sw.c_fill_var(planWT_ist_am);
        sw.c_fill_var(planWTS);

        // Raum-Variablen
        for (D_Raum_SAT ra_sat : planR.keySet()) {
            planR.put(ra_sat, sw.c_new_var());
        }

        // Stunden 
        for (int row = 0; row < mySt.size(); row++) {
            D_Stunde_SAT st = mySt.get_obj_with_row(row);
            st.cnf_pre(sw);
        }

    }

    void cnf_post(SatWrapper sw, D_Schule_SAT sc) throws Exception {
        // this: Jeder Stundenblock hat seinen eigenen Tag
        for (int w = 0; w < MAX_W; w++) {
            for (int t = 0; t < MAX_T; t++) {
                LinkedList<Integer> list = new LinkedList<>();
                for (int row = 0; row < mySt.size(); row++) {
                    D_Stunde_SAT st = mySt.get_obj_with_row(row);
                    list.addLast(st.planWT_ist_am[w][t]);
                }
                int root = sw.c_at_most_one_tree(list);
                sw.c_equal(root, planWT_ist_am[w][t]);
            }
        }

        // this: In jeder Zelle kann höchstens 1 Stundenelement liegen
        for (int w = 0; w < MAX_W; w++) {
            for (int t = 0; t < MAX_T; t++) {
                for (int s = 0; s < MAX_S; s++) {
                    LinkedList<Integer> list = new LinkedList<>();
                    for (int row = 0; row < mySt.size(); row++) {
                        D_Stunde_SAT st = mySt.get_obj_with_row(row);
                        list.addLast(st.planWTS[w][t][s]);
                    }
                    int root = sw.c_at_most_one_tree(list);
                    sw.c_equal(root, planWTS[w][t][s]);
                }
            }
        }

        // this: Genau ein Raum muss gewählt werden
        if (!planR.isEmpty()) {
            LinkedList<Integer> listR = new LinkedList<>();
            for (D_Raum_SAT ra_sat : planR.keySet()) {
                int x = planR.get(ra_sat);
                listR.addLast(x);
            }
            sw.c_exactly_generic(listR, 1);
        }

        // Stunden 
        for (int row = 0; row < mySt.size(); row++) {
            D_Stunde_SAT st = mySt.get_obj_with_row(row);
            st.cnf_post(sw, sc);
        }

        // plan
        sw.c_plan(plan, planWTS, MAX_T, MAX_S, false, false, false, false, false, false, false);

        // nicht_aufeinanderfolgend
        if (nicht_aufeinanderfolgend == 1) {
            if (mySt.size() == 2) {
                LinkedList<Integer> list = new LinkedList<>();
                for (int w = 0; w < MAX_W; w++) {
                    for (int t = 0; t < MAX_T; t++) {
                        list.addLast(planWT_ist_am[w][t]);
                    }
                }
                for (int i = 0; i < list.size(); i++) {
                    sw.c_not_both(list.get(0), list.get(1));
                    list.addLast(list.pollFirst());
                }
            }
        }

        // nicht_gleicher_tag_wie
        for (D_Kurs_SAT ku2 : nicht_gleicher_tag_wie) {
            for (int w = 0; w < MAX_W; w++) {
                for (int t = 0; t < MAX_T; t++) {
                    sw.c_not_both(planWT_ist_am[w][t], ku2.planWT_ist_am[w][t]);
                }
            }
        }
    }

    @Override
    public int hashCode() {
        return super.hashCode();
    }

    @Override
    public boolean equals(Object obj
    ) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final D_Kurs_SAT other = (D_Kurs_SAT) obj;
        if (this.id != other.id) {
            return false;
        }
        return true;
    }

}
