package bar.stundenplan;

import java.util.LinkedList;

/**
 * @author Benjamin A. Bartsch
 */
public class D_Lehrkraft_SAT {

    final int MAX_W;
    final int MAX_T;
    final int MAX_S;

    final int id;
    final String kuerzel;
    final String anzeigename;
    final String bemerkung;
    final int max_spring_pro_tag;
    final int max_spring_pro_woche;
    final boolean darf_vertreten;
    final int min_freie_tage;
    final int max_tagesstunden;
    final int mind_x_ab_stunde;
    final int mind_y_ab_stunde;
    final int mind_x_bis_stunde;
    final int mind_y_bis_stunde;
    final String plan;
    final int min_stunden_pro_tag;
    final int max_stunden_folge;

    final int[][][] planWTS; // Woche, Tag, Stunde
    final int[][] planWT_am_tag_unterricht; // Unterricht ist am Tag ...
    final int[][][] planWTS_hier_oder_davor_unterricht; // Woche, Tag, Stunde
    final int[][][] planWTS_hier_oder_danach_unterricht; // Woche, Tag, Stunde
    final int[][][] planWTS_springstunde; // Woche, Tag, Stunde
    final int[][][] planWTS_vertretungsoption; // Woche, Tag, Stunde
    final int[][][] planWTS_sortiert; // Woche, Tag, Stunde (Stunden sind sortiert)

    final LinkedList<D_Kurs_SAT> myKu = new LinkedList<>();
    final LinkedList<D_Kurs_SAT> myKuHosp = new LinkedList<>();

    public D_Lehrkraft_SAT(int MAX_T, int MAX_S, D_Lehrkraft le, D_Schule sc) {
        this.MAX_W = 2;
        this.MAX_T = MAX_T;
        this.MAX_S = MAX_S;

        this.id = le.id;
        this.kuerzel = le.kuerzel;
        this.anzeigename = le.anzeigename;
        this.bemerkung = le.bemerkung;
        this.darf_vertreten = le.darf_vertreten == 1;
        this.min_freie_tage = le.freie_tage;
        this.mind_x_ab_stunde = le.mind_x_ab_stunde;
        this.mind_y_ab_stunde = le.mind_y_ab_stunde;
        this.mind_x_bis_stunde = le.mind_x_bis_stunde;
        this.mind_y_bis_stunde = le.mind_y_bis_stunde;
        this.plan = le.plan;

        this.max_spring_pro_tag = le.spring_pro_tag_check == 0 ? le.spring_pro_tag_wert : sc.max_spring_pro_tag;
        this.max_spring_pro_woche = le.spring_pro_woche_check == 0 ? le.spring_pro_woche_wert : sc.max_spring_pro_woche;
        this.max_tagesstunden = le.tagesstunden_check == 0 ? le.tagesstunden_wert : sc.max_praesenz_pro_tag;
        this.min_stunden_pro_tag = le.min_stunden_check == 0 ? le.min_stunden_wert : sc.min_stunden_le;
        this.max_stunden_folge = le.max_stunden_folge_check == 0 ? le.max_stunden_folge_wert : sc.max_stunden_folge_le;

        this.planWTS = new int[MAX_W][MAX_T][MAX_S];
        this.planWTS_hier_oder_davor_unterricht = new int[MAX_W][MAX_T][MAX_S];
        this.planWTS_hier_oder_danach_unterricht = new int[MAX_W][MAX_T][MAX_S];
        this.planWTS_springstunde = new int[MAX_W][MAX_T][MAX_S];
        this.planWT_am_tag_unterricht = new int[MAX_W][MAX_T];
        this.planWTS_vertretungsoption = new int[MAX_W][MAX_T][MAX_S];
        this.planWTS_sortiert = new int[MAX_W][MAX_T][MAX_S];
    }

    void cnf_pre(SatWrapper sw) {
        // this
        sw.c_fill_var(planWTS);
        sw.c_fill_var(planWT_am_tag_unterricht);
        sw.c_fill_var(planWTS_hier_oder_davor_unterricht);
        sw.c_fill_var(planWTS_hier_oder_danach_unterricht);
        sw.c_fill_var(planWTS_springstunde);
        sw.c_fill_var(planWTS_vertretungsoption);
        sw.c_fill_var(planWTS_sortiert);
    }

    void cnf_post(SatWrapper sw, D_Berechnungen_SAT bea_sat) throws Exception {
        // this: planWT_am_tag_unterricht
        for (int w = 0; w < MAX_W; w++) {
            for (int t = 0; t < MAX_T; t++) {
                LinkedList<Integer> list = new LinkedList<>();
                for (int s = 0; s < MAX_S; s++) {
                    list.addLast(planWTS[w][t][s]);
                }
                int root = sw.c_OR_tree(list);
                sw.c_equal(root, planWT_am_tag_unterricht[w][t]);
            }
        }
        // this: planWTS
        // this: planWTS_hier_oder_davor_unterricht
        // this: planWTS_hier_oder_danach_unterricht
        // this: planWTS_springstunde
        for (int w = 0; w < MAX_W; w++) {
            for (int t = 0; t < MAX_T; t++) {
                for (int s = 0; s < MAX_S; s++) {
                    // planWTS
                    LinkedList<Integer> list = new LinkedList<>();
                    for (D_Kurs_SAT ku_sat : myKu) {
                        list.addLast(ku_sat.planWTS[w][t][s]);
                    }
                    for (D_Kurs_SAT ku_sat : myKuHosp) {
                        list.addLast(ku_sat.planWTS[w][t][s]);
                    }
                    int root = sw.c_at_most_one_tree(list);
                    sw.c_equal(root, planWTS[w][t][s]);
                    // planWTS_hier_oder_davor_unterricht
                    if (s == 0) {
                        sw.c_equal(planWTS_hier_oder_davor_unterricht[w][t][s], planWTS[w][t][s]);
                    } else {
                        int x = planWTS_hier_oder_davor_unterricht[w][t][s - 1];
                        int y = planWTS[w][t][s];
                        sw.c_equal(planWTS_hier_oder_davor_unterricht[w][t][s], sw.c_new_var_OR(x, y));
                    }
                    // planWTS_hier_oder_danach_unterricht
                    if (s == MAX_S - 1) {
                        sw.c_equal(planWTS_hier_oder_danach_unterricht[w][t][s], planWTS[w][t][s]);
                    } else {
                        int x = planWTS_hier_oder_danach_unterricht[w][t][s + 1];
                        int y = planWTS[w][t][s];
                        sw.c_equal(planWTS_hier_oder_danach_unterricht[w][t][s], sw.c_new_var_OR(x, y));
                    }
                    // planWTS_springstunde
                    if ((s == 0) || (s == MAX_S - 1)) {
                        sw.c_equal(planWTS_springstunde[w][t][s], sw.c_FALSE());
                    } else {
                        int x = planWTS_hier_oder_davor_unterricht[w][t][s - 1];
                        int y = -planWTS[w][t][s];
                        int z = planWTS_hier_oder_danach_unterricht[w][t][s + 1];
                        sw.c_equal(planWTS_springstunde[w][t][s], sw.c_new_var_AND(x, sw.c_new_var_AND(y, z)));
                    }
                }
            }
        }
        // this: max_spring_tag
        // this: max_spring_woche
        for (int w = 0; w < MAX_W; w++) {
            LinkedList<Integer> listWoche = new LinkedList<>();
            for (int t = 0; t < MAX_T; t++) {
                LinkedList<Integer> listTag = new LinkedList<>();
                for (int s = 0; s < MAX_S; s++) {
                    listTag.addLast(planWTS_springstunde[w][t][s]);
                    listWoche.addLast(planWTS_springstunde[w][t][s]);
                }
                if (!bea_sat.deaktiviere_le_springstunden_tag) {
                    sw.c_at_most_generic(listTag, max_spring_pro_tag);
                }
            }
            if (!bea_sat.deaktiviere_le_springstunden_woche) {
                sw.c_at_most_generic(listWoche, max_spring_pro_woche);
            }
        }

        // this: min_freie_tage
        if ((min_freie_tage > 0) && (!bea_sat.deaktiviere_le_freie_tage)) {
            for (int w = 0; w < MAX_W; w++) {
                LinkedList<Integer> listFreieTage = new LinkedList<>();
                for (int t = 0; t < MAX_T; t++) {
                    listFreieTage.addLast(-planWT_am_tag_unterricht[w][t]);
                }
                sw.c_at_least_generic(listFreieTage, min_freie_tage);
            }
        }

        // this: max_praesenz_pro_tag
        if (!bea_sat.deaktiviere_le_tagesstundenpraesenz) {
            for (int w = 0; w < MAX_W; w++) {
                for (int t = 0; t < MAX_T; t++) {
                    for (int s1 = 0, s2 = s1 + max_tagesstunden; s2 < MAX_S; s1++, s2++) {
                        int x = planWTS_hier_oder_davor_unterricht[w][t][s1];
                        int y = planWTS_hier_oder_danach_unterricht[w][t][s2];
                        sw.c_implication(x, -y);
                    }
                }
            }
        }

        // this: mind_x_ab_stunde;
        // this: mind_y_ab_stunde;
        if ((mind_x_ab_stunde > 0) && (mind_y_ab_stunde >= 2) && (!bea_sat.deaktiviere_le_erst_ab_stunde)) {
            for (int w = 0; w < MAX_W; w++) {
                LinkedList<Integer> listMindestensAb = new LinkedList<>();
                for (int t = 0; t < MAX_T; t++) {
                    int s = mind_y_ab_stunde - 2;
                    int x = -planWTS_hier_oder_davor_unterricht[w][t][s]; // 0, 1, [2], 3, 4 
                    listMindestensAb.addLast(x);
                }
                sw.c_at_least_generic(listMindestensAb, mind_x_ab_stunde);
            }
        }

        // this: mind_x_bis_stunde;
        // this: mind_y_bis_stunde;
        if ((mind_x_bis_stunde > 0) && (mind_y_bis_stunde < MAX_S) && (!bea_sat.deaktiviere_le_nur_bis_stunde)) {
            for (int w = 0; w < MAX_W; w++) {
                LinkedList<Integer> listMindestensBis = new LinkedList<>();
                for (int t = 0; t < MAX_T; t++) {
                    int s = mind_y_bis_stunde;
                    int x = -planWTS_hier_oder_danach_unterricht[w][t][s];
                    listMindestensBis.addLast(x);
                }
                sw.c_at_least_generic(listMindestensBis, mind_x_bis_stunde);
            }
        }

        // plan
        sw.c_plan(plan, planWTS, MAX_T, MAX_S
                , bea_sat.deaktiviere_le_plan_spalte_min, bea_sat.deaktiviere_le_plan_spalte_max
                , bea_sat.deaktiviere_le_plan_zeile_min, bea_sat.deaktiviere_le_plan_zeile_max
                , bea_sat.deaktiviere_le_plan_zeile_min_ab, bea_sat.deaktiviere_le_plan_zeile_max_ab, bea_sat.deaktiviere_le_plan_zellen);

        // planWTS_sortiert
        for (int w = 0; w < MAX_W; w++) {
            for (int t = 0; t < MAX_T; t++) {
                // Sortiere s
                LinkedList<Integer> sortiert_s = new LinkedList<>();
                for (int s = 0; s < MAX_S; s++) {
                    sortiert_s.addLast(planWTS[w][t][s]);
                }
                sw.c_bitonic_sort(sortiert_s);
                // equal
                for (int s = 0; s < MAX_S; s++) {
                    sw.c_equal(planWTS_sortiert[w][t][s], sortiert_s.get(s));
                }
            }
        }

        // planWTS_vertretungsoption
        for (int w = 0; w < MAX_W; w++) {
            for (int t = 0; t < MAX_T; t++) {
                for (int s = 0; s < MAX_S; s++) {
                    if (darf_vertreten) {
                        int a2 = s - 2 >= 0 ? planWTS[w][t][s - 2] : sw.c_TRUE(); // davor2
                        int a1 = s - 1 >= 0 ? planWTS[w][t][s - 1] : sw.c_TRUE(); // davor1
                        int b = -planWTS[w][t][s]; // Lücke
                        int c1 = s + 1 < MAX_S ? planWTS[w][t][s + 1] : sw.c_TRUE(); // danach1
                        int c2 = s + 2 < MAX_S ? planWTS[w][t][s + 2] : sw.c_TRUE(); // danach2

                        int a = sw.c_new_var_OR(a1, a2); // davor
                        int c = sw.c_new_var_OR(c1, c2); // danach
                        int ac = sw.c_new_var_AND(a, c); // davor UND danach
                        int d = 6 < MAX_S ? -planWTS_sortiert[w][t][6] : sw.c_TRUE(); // 1111110 --> Maximal 6. Stunden Unterricht (ohne Vertretung)
                        int bd = sw.c_new_var_AND(b, d); // Lücke UND maximal 6. Stunden
                        int abcd = sw.c_new_var_AND(ac, bd); // davor UND danach UND Lücke UND maximal 6. Stunden
                        sw.c_equal(planWTS_vertretungsoption[w][t][s], abcd);
                    } else {
                        sw.c_1(-planWTS_vertretungsoption[w][t][s]);
                    }
                }
            }
        }

        // lage eines stundenelementes eines lehrers (gilt nicht für hospitations-kurse)
        for (D_Kurs_SAT ku_sat : myKu) {
            for (D_Stunde_SAT st_sat : ku_sat.mySt.rows()) {
                if (st_sat.lage_le != Config.STUNDEN_LAGE_EGAL) {
                    st_sat.cnf_post_plan_lage(sw, st_sat.lage_le, planWTS_hier_oder_davor_unterricht, planWTS_hier_oder_danach_unterricht);
                }
            }
        }

        // min_stunden_pro_tag
        if ((min_stunden_pro_tag >= 2) && (!bea_sat.deaktiviere_le_min_stunden_pro_tag)) {
            for (int w = 0; w < MAX_W; w++) {
                for (int t = 0; t < MAX_T; t++) {
                    // freier tag ODER mindestanzahl
                    int x = -planWT_am_tag_unterricht[w][t]; // freier tag
                    int y = sw.c_FALSE(); // mindestanzahl
                    if (min_stunden_pro_tag <= planWTS_sortiert[w][t].length) {
                        y = planWTS_sortiert[w][t][min_stunden_pro_tag - 1];
                    }
                    sw.c_2(x, y); // ODER
                }
            }
        }

        // max_stunden_folge
        if ((max_stunden_folge > 0) && (max_stunden_folge < 16) && (!bea_sat.deaktiviere_le_max_stunden_in_folge)) {
            for (int w = 0; w < MAX_W; w++) {
                for (int t = 0; t < MAX_T; t++) {
                    for (int s1 = 0, s2 = max_stunden_folge; s2 < MAX_S; s1++, s2++) {
                        // Verbiete 'max_stunden_folge' in Folge (oder mehr)
                        LinkedList<Integer> list = new LinkedList<>();
                        for (int s = s1; s <= s2; s++) {
                            list.addLast(planWTS[w][t][s]);
                        }
                        int root = sw.c_AND_tree(list);
                        sw.c_1(-root);
                    }
                }
            }
        }

    }

    @Override
    public int hashCode() {
        return id;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final D_Lehrkraft_SAT other = (D_Lehrkraft_SAT) obj;
        if (this.id != other.id) {
            return false;
        }
        return true;
    }

    void println_debug2(SatWrapper sw) {
        if (kuerzel.equals("EHAS")) {
            System.out.println("" + kuerzel);

            System.out.println("Woche " + (0 + 1));
            for (int s = 0; s < MAX_S; s++) {
                System.out.print("" + (s + 1) + ": ");
                for (int t = 0; t < MAX_T; t++) {
                    int wert = sw.is_true(planWTS_sortiert[0][t][s]) ? 1 : 0;
                    System.out.print(" " + wert);
                }
                System.out.println("");
            }

            System.out.println("Woche " + (0 + 1));
            for (int s = 0; s < MAX_S; s++) {
                System.out.print("" + (s + 1) + ": ");
                for (int t = 0; t < MAX_T; t++) {
                    int wert = sw.is_true(planWTS_springstunde[0][t][s]) ? 1 : 0;
                    System.out.print(" " + wert);
                }
                System.out.println("");
            }

            System.out.println("Woche " + (0 + 1));
            for (int s = 0; s < MAX_S; s++) {
                System.out.print("" + (s + 1) + ": ");
                for (int t = 0; t < MAX_T; t++) {
                    int wert = sw.is_true(planWTS_vertretungsoption[0][t][s]) ? 1 : 0;
                    System.out.print(" " + wert);
                }
                System.out.println("");
            }

        }
    }

}
