package bar.stundenplan;

import java.util.LinkedList;
import java.util.TreeMap;
import javax.swing.JTextPane;

/**
 * @author Benjamin A. Bartsch
 */
public class D_Schule_SOL {

    final int MAX_W;

    final int MAX_T;
    final int MAX_S;
    final int status;
    final long millis;

    final RowMap<D_Kurs_SOL> myKu = new RowMap<>();
    final RowMap<D_Lehrkraft_SOL> myLe = new RowMap<>();
    final RowMap<D_Klasse_SOL> myKl = new RowMap<>();
    final RowMap<D_Fach_SOL> myFa = new RowMap<>();
    final RowMap<D_Raum_SOL> myRa = new RowMap<>();
    final RowMap<D_Kopplung_SOL> myKo = new RowMap<>();

    public D_Schule_SOL(int MAX_W, int MAX_T, int MAX_S, int status, long millis) {
        this.MAX_W = MAX_W;
        this.MAX_T = MAX_T;
        this.MAX_S = MAX_S;
        this.status = status;
        this.millis = millis;
    }

    void html_erzeuge_alles(JTextPane beTextPaneAlles, JTextPane beTextPaneStatistik, JTextPane beTextPaneLehrkraefte, JTextPane beTextPaneKlassen, JTextPane beTextPaneRaeume, JTextPane beTextPaneVertretungsreserven, JTextPane beTextPaneVertretungsoptionen) {
        // lists
        LinkedList<String> htmlAlles = new LinkedList<>();
        LinkedList<String> htmlStatistik = new LinkedList<>();
        LinkedList<String> htmlLehrkraefte = new LinkedList<>();
        LinkedList<String> htmlKlassen = new LinkedList<>();
        LinkedList<String> htmlRaeume = new LinkedList<>();
        LinkedList<String> htmlVertretungsreserven = new LinkedList<>();
        LinkedList<String> htmlVertretungsoptionen = new LinkedList<>();

        // erzeugen
        html_erzeuge_statistik(htmlStatistik);
        html_erzeuge_statistik(htmlAlles);
        html_erzeuge_lehrkraefte(htmlLehrkraefte);
        html_erzeuge_lehrkraefte(htmlAlles);
        html_erzeuge_klassen(htmlKlassen);
        html_erzeuge_klassen(htmlAlles);
        html_erzeuge_raeume(htmlRaeume);
        html_erzeuge_raeume(htmlAlles);
        html_erzeuge_vertretungsreserven(htmlVertretungsreserven);
        html_erzeuge_vertretungsreserven(htmlAlles);
        html_erzeuge_vertretungsoptionen(htmlVertretungsoptionen);
        html_erzeuge_vertretungsoptionen(htmlAlles);

        // finish
        html_finish_to(htmlAlles, beTextPaneAlles);
        html_finish_to(htmlStatistik, beTextPaneStatistik);
        html_finish_to(htmlLehrkraefte, beTextPaneLehrkraefte);
        html_finish_to(htmlKlassen, beTextPaneKlassen);
        html_finish_to(htmlRaeume, beTextPaneRaeume);
        html_finish_to(htmlVertretungsreserven, beTextPaneVertretungsreserven);
        html_finish_to(htmlVertretungsoptionen, beTextPaneVertretungsoptionen);
    }

    private void html_finish_to(LinkedList<String> html, JTextPane tp) {
        // html pre
        html.addFirst("<body>");
        html.addFirst("<html>");
        html.addFirst("<!DOCTYPE html>");

        // html post
        html.addLast("</body>");
        html.addLast("</html>");

        // html escape
        int size = html.size();
        for (int i = 0; i < size; i++) {
            String s = html.pollFirst();
            s = s.replaceAll("Ä", "&Auml;");
            s = s.replaceAll("ä", "&auml;");
            s = s.replaceAll("Ö", "&Ouml;");
            s = s.replaceAll("ö", "&ouml;");
            s = s.replaceAll("Ü", "&Uuml;");
            s = s.replaceAll("ü", "&uuml;");
            s = s.replaceAll("ß", "&szlig;");
            html.addLast(s);
        }

        // html --> string
        StringBuilder sb = new StringBuilder();
        for (String line : html) {
            sb.append(line);
        }
        String sHTML = sb.toString();

        // string --> textpane
        tp.setText(sHTML);
    }

    private void html_erzeuge_statistik(LinkedList<String> html) {

        html.addLast("<h1>Statistik</h1>");

        if (status == Config.SOLVER_UNKNOWN) {
            html.addLast("Berechnung läuft ...<br>");
            return;
        }

        html.addLast("Zeitbedarf " + ConfigHTML.millisToString(millis) + " : <b>" + (status == Config.SOLVER_SAT ? "GELÖST" : "NICHT LÖSBAR!!!") + "</b><br>");
        html.addLast("Wochentage = " + MAX_T + ", Stunden pro Tag = " + MAX_S + ", Kurse = " + myKu.size() + ", Lehrkräfte = " + myLe.size() + ", Klassen = " + myKl.size() + ", Fächer = " + myFa.size() + ", Räume = " + myRa.size() + ", Kopplungen = " + myKo.size() + "<br>");

        if (status == Config.SOLVER_UNSAT) {
            html.addLast("Die aktuellen Einstellungen sind NICHT nicht lösbar!<br>");
            return;
        }

        // erzeuge
        html.addLast("<h2>Lehrkräfte - Tage mit nur einer einzigen Stunde</h2>");
        TreeMap<Integer, LinkedList<String>> mapEinzel = new TreeMap<>();
        for (int i = 0; i < myLe.size(); i++) {
            D_Lehrkraft_SOL le_sol = myLe.get_obj_with_row(i);
            int summe = le_sol.gib_einzelstunden(MAX_W, MAX_T, MAX_S);
            if (summe > 0) {
                if (!mapEinzel.containsKey(summe)) {
                    mapEinzel.put(summe, new LinkedList<>());
                }
                mapEinzel.get(summe).addLast(le_sol.kuerzel);
            }
        }
        for (int summe : mapEinzel.descendingKeySet()) {
            html.addLast(((summe + 0.0) / MAX_W) + ":  " + mapEinzel.get(summe) + "<br>");
        }

        html.addLast("<h2>Lehrkräfte - Größte Stundenanzahl in Folge</h2>");
        TreeMap<Integer, LinkedList<String>> mapStundenFolge = new TreeMap<>();
        for (int i = 0; i < myLe.size(); i++) {
            D_Lehrkraft_SOL le_sol = myLe.get_obj_with_row(i);
            int std = le_sol.gib_groesste_stunden_folge(MAX_W, MAX_T, MAX_S);
            if (!mapStundenFolge.containsKey(std)) {
                mapStundenFolge.put(std, new LinkedList<>());
            }
            mapStundenFolge.get(std).addLast(le_sol.kuerzel);
        }
        for (int std : mapStundenFolge.descendingKeySet()) {
            html.addLast((std) + ":  " + mapStundenFolge.get(std) + "<br>");
        }

        html.addLast("<h2>Lehrkräfte - Größte Tagespräsenz (Alle Stunden am Tag inklusive Springstunden)</h2>");
        TreeMap<Integer, LinkedList<String>> mapStundenPraesenz = new TreeMap<>();
        for (int i = 0; i < myLe.size(); i++) {
            D_Lehrkraft_SOL le_sol = myLe.get_obj_with_row(i);
            int std = le_sol.gib_groesste_praesenz_folge(MAX_W, MAX_T, MAX_S);
            if (!mapStundenPraesenz.containsKey(std)) {
                mapStundenPraesenz.put(std, new LinkedList<>());
            }
            mapStundenPraesenz.get(std).addLast(le_sol.kuerzel);
        }
        for (int std : mapStundenPraesenz.descendingKeySet()) {
            html.addLast((std) + ":  " + mapStundenPraesenz.get(std) + "<br>");
        }

        html.addLast("<h2>Lehrkräfte - Maximale Anzahl an Springstunden (pro Tag)</h2>");
        TreeMap<Integer, LinkedList<String>> mapSpringTag = new TreeMap<>();
        for (int i = 0; i < myLe.size(); i++) {
            D_Lehrkraft_SOL le_sol = myLe.get_obj_with_row(i);
            int max = le_sol.gib_groesste_springstunden_tag(MAX_W, MAX_T, MAX_S);
            if (!mapSpringTag.containsKey(max)) {
                mapSpringTag.put(max, new LinkedList<>());
            }
            mapSpringTag.get(max).addLast(le_sol.kuerzel);
        }
        for (int max : mapSpringTag.descendingKeySet()) {
            html.addLast((max) + ":  " + mapSpringTag.get(max) + "<br>");
        }

        html.addLast("<h2>Lehrkräfte - Maximale Anzahl an Springstunden (pro Woche)</h2>");
        TreeMap<Integer, LinkedList<String>> mapSpringWoche = new TreeMap<>();
        for (int i = 0; i < myLe.size(); i++) {
            D_Lehrkraft_SOL le_sol = myLe.get_obj_with_row(i);
            int max = le_sol.gib_groesste_springstunden_woche(MAX_W, MAX_T, MAX_S);
            if (!mapSpringWoche.containsKey(max)) {
                mapSpringWoche.put(max, new LinkedList<>());
            }
            mapSpringWoche.get(max).addLast(le_sol.kuerzel);
        }
        for (int max : mapSpringWoche.descendingKeySet()) {
            html.addLast(((max + 0.0) / MAX_W) + ":  " + mapSpringWoche.get(max) + "<br>");
        }

        html.addLast("<h2>Kurse - Aufeinanderfolgende (nur " + 2 + "-tägige)</h2>");
        int summeAufeinanderfolgend = 0;
        for (int i = 0; i < myKu.size(); i++) {
            D_Kurs_SOL ku_sol = myKu.get_obj_with_row(i);
            if (ku_sol.ist_aufeinanderfolgend(MAX_W, MAX_T, MAX_S, 2)) {
                summeAufeinanderfolgend++;
            }
        }
        html.addLast("Summe = " + summeAufeinanderfolgend + " <br>");
        for (int i = 0; i < myKu.size(); i++) {
            D_Kurs_SOL ku_sol = myKu.get_obj_with_row(i);
            if (ku_sol.ist_aufeinanderfolgend(MAX_W, MAX_T, MAX_S, 2)) {
                html.addLast(ku_sol.get_name_all() + " <br>");
            }
        }

    }

    private void html_erzeuge_lehrkraefte(LinkedList<String> html) {

        html.addLast("<h1>Lehrkräfte</h1>");

        if (status == Config.SOLVER_UNKNOWN) {
            html.addLast("Berechnung läuft ...<br>");
            return;
        }

        if (status == Config.SOLVER_UNSAT) {
            html.addLast("Die aktuellen Einstellungen sind NICHT nicht lösbar!<br>");
            return;
        }

        for (int row = 0; row < myLe.size(); row++) {
            D_Lehrkraft_SOL le_sol = myLe.get_obj_with_row(row);
            le_sol.erzeuge_ausgabe(html, MAX_W, MAX_T, MAX_S);
        }

    }

    private void html_erzeuge_klassen(LinkedList<String> html) {
        html.addLast("<h1>Klassen</h1>");

        if (status == Config.SOLVER_UNKNOWN) {
            html.addLast("Berechnung läuft ...<br>");
            return;
        }

        if (status == Config.SOLVER_UNSAT) {
            html.addLast("Die aktuellen Einstellungen sind NICHT nicht lösbar!<br>");
            return;
        }

        for (int row = 0; row < myKl.size(); row++) {
            D_Klasse_SOL kl_sol = myKl.get_obj_with_row(row);
            kl_sol.erzeuge_ausgabe(html, MAX_W, MAX_T, MAX_S);
        }

    }

    private void html_erzeuge_raeume(LinkedList<String> html) {
        html.addLast("<h1>Räume</h1>");

        if (status == Config.SOLVER_UNKNOWN) {
            html.addLast("Berechnung läuft ...<br>");
            return;
        }

        if (status == Config.SOLVER_UNSAT) {
            html.addLast("Die aktuellen Einstellungen sind NICHT nicht lösbar!<br>");
            return;
        }

        for (int row = 0; row < myRa.size(); row++) {
            D_Raum_SOL ra_sol = myRa.get_obj_with_row(row);
            ra_sol.erzeuge_ausgabe(html, MAX_W, MAX_T, MAX_S);
        }

    }

    private void html_erzeuge_vertretungsreserven(LinkedList<String> html) {
        html.addLast("<h1>Vertretungsreserven</h1>");

        if (status == Config.SOLVER_UNKNOWN) {
            html.addLast("Berechnung läuft ...<br>");
            return;
        }

        if (status == Config.SOLVER_UNSAT) {
            html.addLast("Die aktuellen Einstellungen sind NICHT nicht lösbar!<br>");
            return;
        }

        // erzeuge wochenmatrix
        String[][][] planS = new String[MAX_S][MAX_T][MAX_W];
        for (int w = 0; w < MAX_W; w++) {
            for (int t = 0; t < MAX_T; t++) {
                for (int s = 0; s < MAX_S; s++) {
                    planS[s][t][w] = "";

                    // summe
                    int summe = 0;
                    for (int i = 0; i < myKu.size(); i++) {
                        summe += myKu.get_obj_with_row(i).get_vertretungsreserven(w, s, t); // TODO auch die Kurse dazu anzeigen?
                    }
                    if (summe > 0) {
                        planS[s][t][w] += " " + summe + "<br>";

                    }

                    // beteiligte kurse
                    for (int i = 0; i < myKu.size(); i++) {
                        D_Kurs_SOL ku_sol = myKu.get_obj_with_row(i);
                        if (ku_sol.vertretungsreserven > 0) {
                            if (ku_sol.hat_unterricht(w, t, s)) {
                                planS[s][t][w] += " " + ku_sol.get_zelle_all(w, t, s) + " (" + ku_sol.vertretungsreserven + ")" + "<br>";
                            }
                        }
                    }

                }
            }
        }
        ConfigHTML.toHtmlTable(MAX_S, MAX_T, MAX_W, planS, html);

    }

    private void html_erzeuge_vertretungsoptionen(LinkedList<String> html) {
        html.addLast("<h1>Vertretungsoptionen</h1>");

        if (status == Config.SOLVER_UNKNOWN) {
            html.addLast("Berechnung läuft ...<br>");
            return;
        }

        if (status == Config.SOLVER_UNSAT) {
            html.addLast("Die aktuellen Einstellungen sind NICHT nicht lösbar!<br>");
            return;
        }

        // erzeuge wochenmatrix
        String[][][] planS = new String[MAX_S][MAX_T][MAX_W];
        for (int w = 0; w < MAX_W; w++) {
            for (int s = 0; s < MAX_S; s++) {
                for (int t = 0; t < MAX_T; t++) {
                    planS[s][t][w] = "";
                    // summe 
                    int summe = 0;
                    for (int i = 0; i < myLe.size(); i++) {
                        D_Lehrkraft_SOL le_sol = myLe.get_obj_with_row(i);
                        summe += le_sol.get_vertretungsoption(w, t, s, MAX_S);
                    }
                    if (summe > 0) {
                        planS[s][t][w] += " " + summe + "<br>";
                    }
                    // beteiligte Lehrkräfte
                    for (int i = 0; i < myLe.size(); i++) {
                        D_Lehrkraft_SOL le_sol = myLe.get_obj_with_row(i);
                        if (le_sol.get_vertretungsoption(w, t, s, MAX_S) > 0) {
                            planS[s][t][w] += " " + le_sol.kuerzel + "<br>";
                        }
                    }
                }
            }
        }
        ConfigHTML.toHtmlTable(MAX_S, MAX_T, MAX_W, planS, html);

    }

}
