package bar.stundenplan;

/**
 * @author Benjamin A. Bartsch
 */
public abstract class D_Stunde_SAT {

    final int MAX_W = 2;
    final int MAX_T;
    final int MAX_S;
    final int id;
    final int stunden;
    final String plan;
    final int lage_kl;
    final int lage_le;

    final int[][][] planWTS; // Woche, Tag, Stunde
    final int[][] planWT_ist_am; // Die Stunde ist am Tag...
    final int[] planS_beginnt; // Die Stunde beginnt in Stunde...
    final int[] planS_endet; // Die Stunde endet in Stunde...
    final int[] planS_ist; // Die Stunde ist in Stunde ...

    public D_Stunde_SAT(int MAX_TAGE, int MAX_STUNDEN, int id, int wochenstunden, String plan, int lage_kl, int lage_le) {
        this.MAX_T = MAX_TAGE;
        this.MAX_S = MAX_STUNDEN;
        this.id = id;
        this.stunden = wochenstunden;
        this.plan = plan;
        this.lage_kl = lage_kl;
        this.lage_le = lage_le;

        this.planWTS = new int[MAX_W][MAX_TAGE][MAX_STUNDEN];
        this.planWT_ist_am = new int[MAX_W][MAX_TAGE];
        this.planS_beginnt = new int[MAX_STUNDEN];
        this.planS_endet = new int[MAX_STUNDEN];
        this.planS_ist = new int[MAX_STUNDEN];
    }

    void cnf_pre(SatWrapper sw) {
        sw.c_fill_var(planWT_ist_am);
        sw.c_fill_var(planWTS);
        sw.c_fill_var(planS_beginnt);
        sw.c_fill_var(planS_endet);
        sw.c_fill_var(planS_ist);
    }

    abstract void cnf_post(SatWrapper sw, D_Schule_SAT sc) throws Exception;

    void cnf_post_planWTS(SatWrapper sw) throws Exception {
        for (int w = 0; w < MAX_W; w++) {
            for (int t = 0; t < MAX_T; t++) {
                int x = planWT_ist_am[w][t];
                for (int s = 0; s < MAX_S; s++) {
                    int y = planS_ist[s];
                    int z = sw.c_new_var_AND(x, y);
                    sw.c_equal(planWTS[w][t][s], z);
                }
            }
        }
    }

    void cnf_post_plan_lage(SatWrapper sw, int lage, int[][][] wts_hier_oder_davor, int[][][] wts_hier_oder_danach) throws Exception {
        for (int w = 0; w < 2; w++) {
            for (int t = 0; t < MAX_T; t++) {
                for (int s = 0; s < MAX_S; s++) {
                    if (lage == Config.STUNDEN_LAGE_AM_ANFANG) {
                        int startet_hier = sw.c_new_var_AND(planS_beginnt[s], planWT_ist_am[w][t]);
                        int davor = (s == 0) ? sw.c_FALSE() : wts_hier_oder_davor[w][t][s - 1];
                        sw.c_implication(startet_hier, -davor);
                    }
                    if (lage == Config.STUNDEN_LAGE_NICHT_AM_ANFANG) {
                        int startet_hier = sw.c_new_var_AND(planS_beginnt[s], planWT_ist_am[w][t]);
                        int davor = (s == 0) ? sw.c_FALSE() : wts_hier_oder_davor[w][t][s - 1];
                        sw.c_implication(startet_hier, +davor);
                    }
                    if (lage == Config.STUNDEN_LAGE_AM_ENDE) {
                        int endet_hier = sw.c_new_var_AND(planS_endet[s], planWT_ist_am[w][t]);
                        int danach = (s + 1 >= MAX_S) ? sw.c_FALSE() : wts_hier_oder_danach[w][t][s + 1];
                        sw.c_implication(endet_hier, -danach);
                    }
                    if (lage == Config.STUNDEN_LAGE_NICHT_AM_ENDE) {
                        int endet_hier = sw.c_new_var_AND(planS_endet[s], planWT_ist_am[w][t]);
                        int danach = (s + 1 >= MAX_S) ? sw.c_FALSE() : wts_hier_oder_danach[w][t][s + 1];
                        sw.c_implication(endet_hier, +danach);
                    }
                    if (lage == Config.STUNDEN_LAGE_AM_RAND) {
                        for (int s2 = s; s2 < MAX_S; s2++) {
                            int startet_hier = sw.c_new_var_AND(planS_beginnt[s], planWT_ist_am[w][t]);
                            int davor = (s == 0) ? sw.c_FALSE() : wts_hier_oder_davor[w][t][s - 1];
                            int endet_hier = sw.c_new_var_AND(planS_endet[s2], planWT_ist_am[w][t]);
                            int danach = (s2 + 1 >= MAX_S) ? sw.c_FALSE() : wts_hier_oder_danach[w][t][s2 + 1];
                            int startet_und_endet_hier = sw.c_new_var_AND(startet_hier, endet_hier);
                            int am_rand = sw.c_new_var_OR(-davor, -danach);
                            sw.c_implication(startet_und_endet_hier, am_rand);
                        }
                    }
                    if (lage == Config.STUNDEN_LAGE_NICHT_AM_RAND) {
                        for (int s2 = s; s2 < MAX_S; s2++) {
                            int startet_hier = sw.c_new_var_AND(planS_beginnt[s], planWT_ist_am[w][t]);
                            int davor = (s == 0) ? sw.c_FALSE() : wts_hier_oder_davor[w][t][s - 1];
                            int endet_hier = sw.c_new_var_AND(planS_endet[s2], planWT_ist_am[w][t]);
                            int danach = (s2 + 1 >= MAX_S) ? sw.c_FALSE() : wts_hier_oder_danach[w][t][s2 + 1];
                            int startet_und_endet_hier = sw.c_new_var_AND(startet_hier, endet_hier);
                            int am_rand = sw.c_new_var_OR(-davor, -danach);
                            sw.c_implication(startet_und_endet_hier, -am_rand);
                        }
                    }
                }
            }
        }

    }

    void cnf_post_planS(SatWrapper sw, int stundenanzahl) throws Exception {
        // planS_beginnt
        // planS_endet
        sw.c_exactly_generic(planS_beginnt, 1);
        sw.c_exactly_generic(planS_endet, 1);
        for (int s = 0; s < MAX_S; s++) {
            if (s + stundenanzahl - 1 >= MAX_S) {
                sw.c_1(-planS_beginnt[s]);
                continue;
            }
            if (s - stundenanzahl + 1 < 0) {
                sw.c_1(-planS_endet[s]);
                continue;
            }
            sw.c_equal(planS_beginnt[s], planS_endet[s + stundenanzahl - 1]);
        }

        for (int i = 1; i < stundenanzahl; i++) {
            int x = planS_beginnt[MAX_S - i];
            sw.c_1(-x);
        }

        // planS_ist
        for (int s = 0; s < MAX_S; s++) {
            int root = sw.c_OR_array_rueckwaerts(planS_beginnt, s, stundenanzahl);
            sw.c_equal(root, planS_ist[s]);
        }

        // Plan
        sw.c_plan(plan, planWTS, MAX_T, MAX_S, false, false, false, false, false, false, false);
    }

    void cnf_post_planS_verbiete_uebergang(SatWrapper sw, D_Schule_SAT sc) throws Exception {
        if ((sc.keinen_uebergang_1_2) && (planS_ist.length >= 2)) {
            sw.c_not_both(planS_ist[0], planS_ist[1]);
        }
        if ((sc.keinen_uebergang_2_3) && (planS_ist.length >= 3)) {
            sw.c_not_both(planS_ist[1], planS_ist[2]);
        }
        if ((sc.keinen_uebergang_3_4) && (planS_ist.length >= 4)) {
            sw.c_not_both(planS_ist[2], planS_ist[3]);
        }
        if ((sc.keinen_uebergang_4_5) && (planS_ist.length >= 5)) {
            sw.c_not_both(planS_ist[3], planS_ist[4]);
        }
        if ((sc.keinen_uebergang_5_6) && (planS_ist.length >= 6)) {
            sw.c_not_both(planS_ist[4], planS_ist[5]);
        }
        if ((sc.keinen_uebergang_6_7) && (planS_ist.length >= 7)) {
            sw.c_not_both(planS_ist[5], planS_ist[6]);
        }
        if ((sc.keinen_uebergang_7_8) && (planS_ist.length >= 8)) {
            sw.c_not_both(planS_ist[6], planS_ist[7]);
        }
    }

    @Override
    public int hashCode() {
        return id;
    }

    @Override
    public boolean equals(Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        final D_Stunde_SAT other = (D_Stunde_SAT) obj;
        if (this.id != other.id) {
            return false;
        }
        return true;
    }

    
}
