package bar.stundenplan;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.TreeSet;
import javax.swing.JTextArea;
import javax.swing.SwingUtilities;
import javax.swing.event.TreeModelEvent;
import javax.swing.event.TreeModelListener;
import javax.swing.table.DefaultTableModel;
import javax.swing.tree.TreeModel;
import javax.swing.tree.TreePath;

/**
 * @author Benjamin A. Bartsch
 */
public class Model implements TreeModel {

    private static final String[] SAVE_TABLE_ORDER = new String[]{
        "SCHULE", "BERECHNUNG_SCHULE", "FACH", "BERECHNUNG_FACH", "KLASSE",
        "BERECHNUNG_KLASSE", "KOPPLUNG", "BERECHNUNG_KOPPLUNG", "KURS",
        "BERECHNUNG_KURS", "LEHRKRAFT", "BERECHNUNG_LEHRKRAFT", "RAUM", "BERECHNUNG_RAUM",
        "KURS_FACH", "BERECHNUNG_KURS_FACH", "KURS_KLASSE", "BERECHNUNG_KURS_KLASSE",
        "KURS_KOPPLUNG", "BERECHNUNG_KURS_KOPPLUNG", "KURS_LEHRKRAFT", "BERECHNUNG_KURS_LEHRKRAFT",
        "KURS_RAUM", "BERECHNUNG_KURS_RAUM", "STUNDE", "BERECHNUNG_STUNDE", "STUNDE_KO", "BERECHNUNG_STUNDE_KO",
        "LEHRKRAEFTE", "KLASSEN", "RAEUME", "KOPPLUNGEN", "KURSE", "KURS_KURS", "KURS_HOSP", "BERECHNUNG_KURS_HOSP",
        "BERECHNUNGEN"
    };

    private Connection con;

    private final Gui gui;
    private final LinkedList<TreeModelListener> listener;
    private final D_Einstellungen d_einstellungen;
    private final D_Schule d_schule;
    private final D_Lehrkraefte d_lehrkraefte;
    private final D_Klassen d_klassen;
    private final D_Faecher d_faecher;
    private final D_Raeume d_raeume;
    private final D_Kopplungen d_kopplungen;
    private final D_Kurse d_kurse;
    private final D_Berechnungen d_berechnungen;
    private final LinkedList<Object> children_of_root;
    private final LinkedList<Object> children_of_schule;
    private final RowMap<D_Lehrkraft> map_le = new RowMap<>(); // hashcode implemented
    private final RowMap<D_Klasse> map_kl = new RowMap<>(); // hashcode implemented
    private final RowMap<D_Fach> map_fa = new RowMap<>(); // hashcode implemented
    private final RowMap<D_Raum> map_ra = new RowMap<>(); // hashcode implemented
    private final RowMap<D_Kopplung> map_ko = new RowMap<>(); // hashcode implemented
    private final RowMap<D_Kurs> map_ku = new RowMap<>(); // hashcode implemented
    private final RowMap<D_Stunde> map_st = new RowMap<>(); // hashcode implemented
    private final RowMap<D_Stunde_Ko> map_st_ko = new RowMap<>(); // hashcode implemented
    private final RowMap<D_Berechnung> map_be = new RowMap<>(); // hashcode implemented
    private final HashMap<Integer, D_Worker> map_process = new HashMap<>();

    final TableModelKursLehrer table_model_kurs_lehrer = new TableModelKursLehrer();
    final TableModelKursKlasse table_model_kurs_klasse = new TableModelKursKlasse();
    final TableModelKursFach table_model_kurs_fach = new TableModelKursFach();
    final TableModelKursRaum table_model_kurs_raum = new TableModelKursRaum();
    final TableModelKursKopplung table_model_kurs_kopplung = new TableModelKursKopplung();
    final TableModelKursKurs table_model_kurs_kurs = new TableModelKursKurs();
    final TableModelKursHosp table_model_kurs_hosp = new TableModelKursHosp();

    private final TreeSet<String> table_with_single_line;

    public Model(Gui p_gui) {
        this.gui = p_gui;
        this.listener = new LinkedList<>();
        this.d_schule = new D_Schule(5, 6, 2, 4);
        this.d_einstellungen = new D_Einstellungen();
        this.d_lehrkraefte = new D_Lehrkraefte(1, "", map_le);
        this.d_klassen = new D_Klassen(1, "", map_kl);
        this.d_faecher = new D_Faecher(map_fa);
        this.d_raeume = new D_Raeume(1, "", map_ra);
        this.d_kopplungen = new D_Kopplungen(1, "", map_ko);
        this.d_kurse = new D_Kurse(1, "", map_ku);
        this.d_berechnungen = new D_Berechnungen(1, map_be);

        this.children_of_root = new LinkedList<>();
        this.children_of_root.addLast(d_einstellungen);
        this.children_of_root.addLast(d_schule);

        this.children_of_schule = new LinkedList<>();
        this.children_of_schule.addLast(d_lehrkraefte);
        this.children_of_schule.addLast(d_klassen);
        this.children_of_schule.addLast(d_faecher);
        this.children_of_schule.addLast(d_raeume);
        this.children_of_schule.addLast(d_kopplungen);
        this.children_of_schule.addLast(d_kurse);
        this.children_of_schule.addLast(d_berechnungen);

        this.table_with_single_line = new TreeSet<>();
        this.table_with_single_line.add("schule");
        this.table_with_single_line.add("lehrkraefte");
        this.table_with_single_line.add("klassen");
        this.table_with_single_line.add("raeume");
        this.table_with_single_line.add("kopplungen");
        this.table_with_single_line.add("kurse");
        this.table_with_single_line.add("berechnungen");
    }

    // (1)
    void db_save_to_file(File file) throws Exception {
        try (FileOutputStream fos = new FileOutputStream(file)) {
            try (BufferedOutputStream bos = new BufferedOutputStream(fos)) {
                try (ObjectOutputStream oos = new ObjectOutputStream(bos)) {
                    long programm_nr = 8590508723626156522l;
                    oos.writeLong(programm_nr);

                    long versions_nr = 1;
                    oos.writeLong(versions_nr);

                    oos.writeInt(SAVE_TABLE_ORDER.length); // Anzahl an Tabellen
                    for (String sTabelle : SAVE_TABLE_ORDER) {
                        // Tabellenname
                        oos.writeUTF(sTabelle);

                        // Speichern einer Tabelle
                        int rows;
                        try (ResultSet rs = executeQuery("SELECT COUNT(*) FROM " + sTabelle + ";")) {
                            rs.next();
                            rows = rs.getInt(1);
                        }
                        // Zeilenanzahl
                        oos.writeInt(rows);
                        try (ResultSet rs = executeQuery("SELECT * FROM " + sTabelle + ";")) {
                            ResultSetMetaData rsmd = rs.getMetaData();
                            // Spaltenanzahl
                            int cols = rsmd.getColumnCount();
                            oos.writeInt(cols);

                            // Spaltennamen
                            String[] spalten = new String[cols];
                            String[] spaltenTyp = new String[cols];
                            for (int c = 0; c < cols; c++) {
                                spalten[c] = rsmd.getColumnName(c + 1);
                                spaltenTyp[c] = rsmd.getColumnTypeName(c + 1);
                            }
                            oos.writeObject(spalten);
                            oos.writeObject(spaltenTyp);
                            // Inhalte
                            Object[][] data = new Object[rows][cols];
                            for (int r = 0; r < rows; r++) {
                                rs.next();
                                for (int c = 0; c < cols; c++) {
                                    data[r][c] = rs.getObject(c + 1);
                                }
                            }
                            oos.writeObject(data);
                        }
                    }
                }
            }
        }
    }

    // (2)
    void db_load_from_file(File file) throws Exception {

        try (FileInputStream fis = new FileInputStream(file)) {
            try (BufferedInputStream bis = new BufferedInputStream(fis)) {
                try (ObjectInputStream ois = new ObjectInputStream(bis)) {

                    long programm_nr = ois.readLong();
                    if (programm_nr != 8590508723626156522l) {
                        throw new Exception("Ungültige Datei!");
                    }

                    long versions_nr = ois.readLong();
                    if (versions_nr > 1) {
                        throw new Exception("Sie müssen die neuste Programmversion verwenden!");
                    }

                    // Starte das Laden der Daten 
                    db_sc_clear();

                    int cTabellen = ois.readInt();

                    for (int i = 0; i < cTabellen; i++) {
                        String sTabelle = ois.readUTF();
                        int rows = ois.readInt();

                        int cols = ois.readInt();

                        String[] col_name = (String[]) ois.readObject();
                        String[] col_type = (String[]) ois.readObject();

                        String sColumns = "(" + col_name[0];
                        for (int c = 1; c < cols; c++) {
                            sColumns = sColumns + ", " + col_name[c];
                        }
                        sColumns = sColumns + ")";

                        Object[][] data = (Object[][]) ois.readObject();
                        for (int r = 0; r < rows; r++) {
                            String extra = col_type[0].equals("VARCHAR") ? "'" : "";
                            String sValues = "(" + extra + data[r][0] + extra;
                            for (int c = 1; c < cols; c++) {
                                extra = col_type[c].equals("VARCHAR") ? "'" : "";
                                sValues = sValues + ", " + extra + data[r][c] + extra;
                            }
                            sValues = sValues + ")";
                            if (table_with_single_line.contains(sTabelle.toLowerCase())) {
                                db_executeUpdate("DELETE FROM " + sTabelle + ";");
                            }
                            db_executeUpdate("INSERT INTO " + sTabelle + " " + sColumns + " VALUES " + sValues);
                        }

                    }
                }
            }
        }

        db_sc_reload(false);
    }

    // (3)
    @Override
    public int getChildCount(Object parent) {
        try {
            if (parent == this) {
                return children_of_root.size();
            }
            if (parent == d_einstellungen) {
                return 0;
            }
            if (parent == d_schule) {
                return children_of_schule.size();
            }
            if (parent == d_lehrkraefte) {
                return map_le.size();
            }
            if (parent == d_klassen) {
                return map_kl.size();
            }
            if (parent == d_faecher) {
                return map_fa.size();
            }
            if (parent == d_raeume) {
                return map_ra.size();
            }
            if (parent == d_kopplungen) {
                return map_ko.size();
            }
            if (parent == d_kurse) {
                return map_ku.size();
            }
            if (parent == d_berechnungen) {
                return map_be.size();
            }
            if (parent instanceof D_Kurs) {
                return ((D_Kurs) parent).st_map.size();
            }
            if (parent instanceof D_Kopplung) {
                return ((D_Kopplung) parent).st_map.size();
            }
            if (parent instanceof D_Lehrkraft) {
                return 0;
            }
            if (parent instanceof D_Klasse) {
                return 0;
            }
            if (parent instanceof D_Fach) {
                return 0;
            }
            if (parent instanceof D_Raum) {
                return 0;
            }

            if (parent instanceof D_Stunde) {
                return 0;
            }
            if (parent instanceof D_Stunde_Ko) {
                return 0;
            }
            if (parent instanceof D_Berechnung) {
                return 0;
            }
            throw new Exception("getChildCount: Unbekannter Knoten '" + parent + "'");
        } catch (Exception ex) {
            handle_exception(ex);
        }
        return 0;
    }

    // (4)
    @Override
    public boolean isLeaf(Object node) {
        try {
            if (node == this) {
                return false;
            }
            if (node == d_einstellungen) {
                return true;
            }
            if (node == d_schule) {
                return children_of_schule.isEmpty();
            }
            if (node == d_lehrkraefte) {
                return map_le.isEmpty();
            }
            if (node == d_klassen) {
                return map_kl.isEmpty();
            }
            if (node == d_faecher) {
                return map_fa.isEmpty();
            }
            if (node == d_raeume) {
                return map_ra.isEmpty();
            }
            if (node == d_kopplungen) {
                return map_ko.isEmpty();
            }
            if (node == d_kurse) {
                return map_ku.isEmpty();
            }
            if (node == d_berechnungen) {
                return map_be.isEmpty();
            }
            if (node instanceof D_Kurs) {
                return ((D_Kurs) node).st_map.isEmpty();
            }
            if (node instanceof D_Kopplung) {
                return ((D_Kopplung) node).st_map.isEmpty();
            }
            if (node instanceof D_Lehrkraft) {
                return true;
            }
            if (node instanceof D_Klasse) {
                return true;
            }
            if (node instanceof D_Fach) {
                return true;
            }
            if (node instanceof D_Raum) {
                return true;
            }
            if (node instanceof D_Stunde) {
                return true;
            }
            if (node instanceof D_Stunde_Ko) {
                return true;
            }
            if (node instanceof D_Berechnung) {
                return true;
            }
            throw new Exception("isLeaf: Unbekannter Knoten '" + node + "'");
        } catch (Exception ex) {
            handle_exception(ex);
        }
        return true;
    }

    // (5)
    @Override
    public Object getChild(Object parent, int index) {
        try {
            if (parent == this) {
                return children_of_root.get(index);
            }
            if (parent == d_schule) {
                return children_of_schule.get(index);
            }
            if (parent == d_lehrkraefte) {
                return map_le.get_obj_with_row(index);
            }
            if (parent == d_klassen) {
                return map_kl.get_obj_with_row(index);
            }
            if (parent == d_faecher) {
                return map_fa.get_obj_with_row(index);
            }
            if (parent == d_raeume) {
                return map_ra.get_obj_with_row(index);
            }
            if (parent == d_kopplungen) {
                return map_ko.get_obj_with_row(index);
            }
            if (parent == d_kurse) {
                return map_ku.get_obj_with_row(index);
            }
            if (parent instanceof D_Kurs) {
                return ((D_Kurs) parent).st_map.get_obj_with_row(index);
            }
            if (parent instanceof D_Kopplung) {
                return ((D_Kopplung) parent).st_map.get_obj_with_row(index);
            }
            if (parent == d_berechnungen) {
                return map_be.get_obj_with_row(index);
            }
            throw new Exception("getChild: Unbekannter Knoten!");
        } catch (Exception ex) {
            handle_exception(ex);
        }
        return "?";
    }

    // (6)
    @Override
    public int getIndexOfChild(Object parent, Object child) {
        try {
            if (parent == this) {
                return children_of_root.indexOf(child);
            }
            if (parent == d_schule) {
                return children_of_schule.indexOf(child);
            }
            if (parent == d_lehrkraefte) {
                return map_le.get_row_of_obj((D_Lehrkraft) child);
            }
            if (parent == d_klassen) {
                return map_kl.get_row_of_obj((D_Klasse) child);
            }
            if (parent == d_faecher) {
                return map_fa.get_row_of_obj((D_Fach) child);
            }
            if (parent == d_raeume) {
                return map_ra.get_row_of_obj((D_Raum) child);
            }
            if (parent == d_kopplungen) {
                return map_ko.get_row_of_obj((D_Kopplung) child);
            }
            if (parent == d_kurse) {
                return map_ku.get_row_of_obj((D_Kurs) child);
            }
            if (parent instanceof D_Kurs) {
                return ((D_Kurs) parent).st_map.get_row_of_obj((D_Stunde) child);
            }
            if (parent instanceof D_Kopplung) {
                return ((D_Kopplung) parent).st_map.get_row_of_obj((D_Stunde_Ko) child);
            }
            if (parent == d_berechnungen) {
                return map_be.get_row_of_obj((D_Berechnung) child);
            }
            throw new Exception("getIndexOfChild: Unbekannter Knoten!");
        } catch (Exception ex) {
            handle_exception(ex);
        }
        return -1;
    }

    public boolean getHasChild(Object parent, Object child) {
        try {
            if (parent == this) {
                return children_of_root.contains(child);
            }
            if (parent == d_schule) {
                return children_of_schule.contains(child);
            }
            if (parent == d_lehrkraefte) {
                return map_le.contains((D_Lehrkraft) child);
            }
            if (parent == d_klassen) {
                return map_kl.contains((D_Klasse) child);
            }
            if (parent == d_faecher) {
                return map_fa.contains((D_Fach) child);
            }
            if (parent == d_raeume) {
                return map_ra.contains((D_Raum) child);
            }
            if (parent == d_kopplungen) {
                return map_ko.contains((D_Kopplung) child);
            }
            if (parent == d_kurse) {
                return map_ku.contains((D_Kurs) child);
            }
            if (parent instanceof D_Kurs) {
                return ((D_Kurs) parent).st_map.contains((D_Stunde) child);
            }
            if (parent instanceof D_Kopplung) {
                return ((D_Kopplung) parent).st_map.contains((D_Stunde_Ko) child);
            }
            if (parent == d_berechnungen) {
                return map_be.contains((D_Berechnung) child);
            }
            throw new Exception("getIndexOfChild: Unbekannter Knoten!");
        } catch (Exception ex) {
            handle_exception(ex);
        }
        return false;
    }

    // (7)
    void db_sc_reload(boolean select_same) throws Exception {

        // Schule erzeugen aus DB
        try (ResultSet rs = executeQuery("SELECT id, max_t, max_s, max_spring_pro_tag, max_spring_pro_woche, max_praesenz_pro_tag,"
                + " plan, keine_einzelstunde_in_1, keine_einzelstunde_in_2, keine_einzelstunde_in_3, keine_einzelstunde_in_4, keine_einzelstunde_in_5, keine_einzelstunde_in_6, planV,"
                + " keinen_uebergang_1_2, keinen_uebergang_2_3, keinen_uebergang_3_4, keinen_uebergang_4_5, keinen_uebergang_5_6, keinen_uebergang_6_7, keinen_uebergang_7_8,"
                + " planO, min_stunden_le, min_aufeinanderfolgend, max_stunden_folge_le FROM schule WHERE id=1;")) {
            for (int row = 0; rs.next(); row++) {
                int id = rs.getInt(1);
                d_schule.max_t = rs.getInt(2);
                d_schule.max_s = rs.getInt(3);
                d_schule.max_spring_pro_tag = rs.getInt(4);
                d_schule.max_spring_pro_woche = rs.getInt(5);
                d_schule.max_praesenz_pro_tag = rs.getInt(6);
                d_schule.plan = rs.getString(7);
                d_schule.keine_einzelstunde_in_1 = rs.getInt(8) == 1;
                d_schule.keine_einzelstunde_in_2 = rs.getInt(9) == 1;
                d_schule.keine_einzelstunde_in_3 = rs.getInt(10) == 1;
                d_schule.keine_einzelstunde_in_4 = rs.getInt(11) == 1;
                d_schule.keine_einzelstunde_in_5 = rs.getInt(12) == 1;
                d_schule.keine_einzelstunde_in_6 = rs.getInt(13) == 1;
                d_schule.planV = rs.getString(14);
                d_schule.keinen_uebergang_1_2 = rs.getInt(15) == 1;
                d_schule.keinen_uebergang_2_3 = rs.getInt(16) == 1;
                d_schule.keinen_uebergang_3_4 = rs.getInt(17) == 1;
                d_schule.keinen_uebergang_4_5 = rs.getInt(18) == 1;
                d_schule.keinen_uebergang_5_6 = rs.getInt(19) == 1;
                d_schule.keinen_uebergang_6_7 = rs.getInt(20) == 1;
                d_schule.keinen_uebergang_7_8 = rs.getInt(21) == 1;
                d_schule.planO = rs.getString(22);
                d_schule.min_stunden_le = rs.getInt(23);
                d_schule.min_aufeinanderfolgend = rs.getInt(24);
                d_schule.max_stunden_folge_le = rs.getInt(25);
            }
        }

        // Schule - Anzahl an Kursen die genau über 2 Tage gehen
        try (ResultSet rs = executeQuery("SELECT COUNT(*) FROM (SELECT kurs_id, COUNT(*) AS anzahl FROM stunde GROUP BY kurs_id HAVING anzahl=2)")) {
            rs.next();
            d_schule.min_aufeinanderfolgend_von = rs.getInt(1);
        }

        // Lehrkräfte erzeugen aus DB
        try (ResultSet rs = executeQuery("SELECT id, plan FROM lehrkraefte WHERE id=1;")) {
            for (int row = 0; rs.next(); row++) {
                d_lehrkraefte.id = rs.getInt(1);
                d_lehrkraefte.plan = rs.getString(2);
            }
        }

        // Jede Lehrkraft aus DB erzeugen
        map_le.clear();
        try (ResultSet rs = executeQuery("SELECT id, kuerzel, anzeigename, bemerkung, darf_vertreten, freie_tage,"
                + " mind_x_ab_stunde, mind_y_ab_stunde, mind_x_bis_stunde, mind_y_bis_stunde,"
                + " spring_pro_tag_check, spring_pro_tag_wert, spring_pro_woche_check, spring_pro_woche_wert,"
                + " tagesstunden_check, tagesstunden_wert, plan, min_stunden_check, min_stunden_wert,"
                + " max_stunden_folge_check, max_stunden_folge_wert"
                + " FROM lehrkraft ORDER BY kuerzel ASC;")) {
            for (int row = 0; rs.next(); row++) {
                int id = rs.getInt(1);
                D_Lehrkraft le = new D_Lehrkraft(id, rs.getString(2), rs.getString(3), rs.getString(4), rs.getInt(5), rs.getInt(6), rs.getInt(7), rs.getInt(8), rs.getInt(9), rs.getInt(10), rs.getInt(11), rs.getInt(12), rs.getInt(13), rs.getInt(14), rs.getInt(15), rs.getInt(16), rs.getString(17), rs.getInt(18), rs.getInt(19), rs.getInt(20), rs.getInt(21));
                map_le.put(id, le);
            }
        }

        // Klassen aus DB erzeugen
        try (ResultSet rs = executeQuery("SELECT id, plan FROM klassen WHERE id=1;")) {
            for (int row = 0; rs.next(); row++) {
                d_klassen.id = rs.getInt(1);
                d_klassen.plan = rs.getString(2);
            }
        }

        // Jede Klasse aus DB erzeugen
        map_kl.clear();
        try (ResultSet rs = executeQuery("SELECT id, kuerzel, plan FROM klasse ORDER BY kuerzel ASC;")) {
            for (int row = 0; rs.next(); row++) {
                int id = rs.getInt(1);
                D_Klasse kl = new D_Klasse(id, rs.getString(2), rs.getString(3));
                map_kl.put(id, kl);
            }
        }

        // Fächer erzeugen aus DB
        map_fa.clear();
        try (ResultSet rs = executeQuery("SELECT id, kuerzel, anzeigename FROM fach ORDER BY kuerzel ASC;")) {
            for (int row = 0; rs.next(); row++) {
                int id = rs.getInt(1);
                D_Fach fa = new D_Fach(id, rs.getString(2), rs.getString(3));
                map_fa.put(id, fa);
            }
        }

        // Räume aus DB erzeugen
        try (ResultSet rs = executeQuery("SELECT id, plan FROM raeume WHERE id=1;")) {
            for (int row = 0; rs.next(); row++) {
                d_raeume.id = rs.getInt(1);
                d_raeume.plan = rs.getString(2);
            }
        }

        // Jeden Raum aus DB erzeugen
        map_ra.clear();
        try (ResultSet rs = executeQuery("SELECT id, kuerzel, plan FROM raum ORDER BY kuerzel ASC;")) {
            for (int row = 0; rs.next(); row++) {
                int id = rs.getInt(1);
                D_Raum ra = new D_Raum(id, rs.getString(2), rs.getString(3));
                map_ra.put(id, ra);
            }
        }

        // Kopplungen aus DB erzeugen
        try (ResultSet rs = executeQuery("SELECT id, plan FROM kopplungen WHERE id=1;")) {
            for (int row = 0; rs.next(); row++) {
                d_kopplungen.id = rs.getInt(1);
                d_kopplungen.plan = rs.getString(2);
            }
        }

        // Jede Kopplung aus DB erzeugen
        map_ko.clear();
        try (ResultSet rs = executeQuery("SELECT id, kuerzel, anzeigename, plan FROM kopplung ORDER BY kuerzel ASC;")) {
            for (int row = 0; rs.next(); row++) {
                int id = rs.getInt(1);
                D_Kopplung ko = new D_Kopplung(id, rs.getString(2), rs.getString(3), rs.getString(4));
                map_ko.put(id, ko);
            }
        }

        // Kurse erzeugen aus DB
        try (ResultSet rs = executeQuery("SELECT id, plan, sortierung FROM kurse WHERE id=1;")) {
            for (int row = 0; rs.next(); row++) {
                d_kurse.id = rs.getInt(1);
                d_kurse.plan = rs.getString(2);
                d_kurse.sortierung = rs.getInt(3);
            }
        }

        // Jeden Kurs aus DB erzeugen
        map_ku.clear();
        TreeSet<Integer> already = new TreeSet<>();
        String sOrderBy = "ORDER BY kurs.id";
        if (d_kurse.sortierung == Config.KURSE_SORTIEREN_NACH_KL_FA_LE) {
            sOrderBy = "ORDER BY klasse.kuerzel ASC, fach.kuerzel ASC, lehrkraft.kuerzel ASC, kurs.id ASC";
        }
        if (d_kurse.sortierung == Config.KURSE_SORTIEREN_NACH_KL_LE_FA) {
            sOrderBy = "ORDER BY klasse.kuerzel ASC, lehrkraft.kuerzel ASC, fach.kuerzel ASC, kurs.id ASC";
        }
        if (d_kurse.sortierung == Config.KURSE_SORTIEREN_NACH_LE_KL_FA) {
            sOrderBy = "ORDER BY lehrkraft.kuerzel ASC, klasse.kuerzel ASC, fach.kuerzel ASC, kurs.id ASC";
        }
        if (d_kurse.sortierung == Config.KURSE_SORTIEREN_NACH_LE_FA_KL) {
            sOrderBy = "ORDER BY lehrkraft.kuerzel ASC, fach.kuerzel ASC, klasse.kuerzel ASC, kurs.id ASC";
        }
        if (d_kurse.sortierung == Config.KURSE_SORTIEREN_NACH_FA_KL_LE) {
            sOrderBy = "ORDER BY fach.kuerzel ASC, klasse.kuerzel ASC, lehrkraft.kuerzel ASC, kurs.id ASC";
        }
        if (d_kurse.sortierung == Config.KURSE_SORTIEREN_NACH_FA_LE_KL) {
            sOrderBy = "ORDER BY fach.kuerzel ASC, lehrkraft.kuerzel ASC, klasse.kuerzel ASC, kurs.id ASC";
        }
        try (ResultSet rs = executeQuery("SELECT kurs.id, kurs.ergaenzung, kurs.plan, kurs.vertretungsreserven, kurs.nicht_aufeinanderfolgend"
                + " FROM kurs"
                + " LEFT JOIN kurs_klasse ON kurs.id=kurs_klasse.kurs_id"
                + " LEFT JOIN klasse ON klasse.id=kurs_klasse.klasse_id "
                + " LEFT JOIN kurs_fach ON kurs.id=kurs_fach.kurs_id"
                + " LEFT JOIN fach ON fach.id=kurs_fach.fach_id "
                + " LEFT JOIN kurs_lehrkraft ON kurs.id=kurs_lehrkraft.kurs_id "
                + " LEFT JOIN lehrkraft ON lehrkraft.id=kurs_lehrkraft.lehrkraft_id "
                + " GROUP BY kurs.id, lehrkraft.id, klasse.id"
                + " " + sOrderBy + ";")) {
            for (int row = 0; rs.next(); row++) {
                int id = rs.getInt(1);
                if (already.add(id)) {
                    D_Kurs ku = new D_Kurs(id, rs.getString(2), rs.getString(3), rs.getInt(4), rs.getInt(5));
                    map_ku.put(id, ku);
                }
            }
        }

        // Stunden der Kurse erzeugen aus DB
        map_st.clear();
        try (ResultSet rs = executeQuery("SELECT id, kurs_id, wochenstunden, typ, plan, lage_kl, lage_le FROM stunde;")) {
            for (int row = 0; rs.next(); row++) {
                int id = rs.getInt(1);
                int wochenstunden = rs.getInt(3);
                D_Stunde st = new D_Stunde(id, rs.getInt(2), wochenstunden, rs.getInt(4), rs.getString(5), rs.getInt(6), rs.getInt(7));
                map_st.put(id, st);
                // Dem Kurs zuordnen
                D_Kurs ku = map_ku.get_obj_with_id(st.ku_id);
                ku.st_map.put(st.id, st);
                ku.stunden += wochenstunden;
            }
        }

        // Stunden der Kopplung erzeugen aus DB
        map_st_ko.clear();
        try (ResultSet rs = executeQuery("SELECT id, kopplung_id, wochenstunden, typ, plan, lage_kl, lage_le FROM stunde_ko;")) {
            for (int row = 0; rs.next(); row++) {
                int id = rs.getInt(1);
                int wochenstunden = rs.getInt(3);
                D_Stunde_Ko st_ko = new D_Stunde_Ko(id, rs.getInt(2), wochenstunden, rs.getInt(4), rs.getString(5), rs.getInt(6), rs.getInt(7));
                map_st_ko.put(id, st_ko);
                // Der Kopplung zuordnen
                D_Kopplung ko = map_ko.get_obj_with_id(st_ko.ko_id);
                ko.st_map.put(st_ko.id, st_ko);
                ko.stunden += wochenstunden;
            }
        }

        // Berechnungen erzeugen aus DB
        map_be.clear();
        try (ResultSet rs = executeQuery("SELECT id, status, millis, max_t, max_s FROM berechnung_schule ORDER BY id DESC;")) {
            for (int row = 0; rs.next(); row++) {
                int id = rs.getInt(1);
                D_Berechnung be = new D_Berechnung(id, rs.getInt(2), rs.getLong(3), rs.getInt(4), rs.getInt(5));
                map_be.put(id, be);
            }
        }

        // kurs_kurs Referenzen zuordnen
        try (ResultSet rs = executeQuery("SELECT kurs_id_1, kurs_id_2 FROM kurs_kurs;")) {
            for (int row = 0; rs.next(); row++) {
                int ku_id_1 = rs.getInt(1);
                int ku_id_2 = rs.getInt(2);
                D_Kurs ku_1 = map_ku.get_obj_with_id(ku_id_1);
                D_Kurs ku_2 = map_ku.get_obj_with_id(ku_id_2);
                ku_1.nicht_gleicher_tag_wie.put(ku_2.id, ku_2); // implicit "ku1/ku2" adds later "ku2/ku1"
            }
        }

        // kurs_lehrkraft Referenzen zuordnen
        try (ResultSet rs = executeQuery("SELECT kurs_id, lehrkraft_id FROM kurs_lehrkraft;")) {
            for (int row = 0; rs.next(); row++) {
                D_Kurs ku = map_ku.get_obj_with_id(rs.getInt(1));
                D_Lehrkraft le = map_le.get_obj_with_id(rs.getInt(2));
                ku.le_list.add(le);
                le.mapKu2.put(ku.id, ku);
                le.stunden += ku.stunden;
            }
        }

        // kurs_hosp Referenzen zuordnen
        try (ResultSet rs = executeQuery("SELECT kurs_id, lehrkraft_id FROM kurs_hosp;")) {
            for (int row = 0; rs.next(); row++) {
                D_Kurs ku = map_ku.get_obj_with_id(rs.getInt(1));
                D_Lehrkraft le = map_le.get_obj_with_id(rs.getInt(2));
                ku.le_list_hosp.add(le);
                le.mapKuHosp2.put(ku.id, ku);
                le.stunden += ku.stunden;
            }
        }

        // kurs_fach Referenzen zuordnen
        try (ResultSet rs = executeQuery("SELECT kurs_id, fach_id FROM kurs_fach;")) {
            for (int row = 0; rs.next(); row++) {
                D_Kurs ku = map_ku.get_obj_with_id(rs.getInt(1));
                D_Fach fa = map_fa.get_obj_with_id(rs.getInt(2));
                ku.fa_list.add(fa);
                fa.mapKu.put(ku.id, ku);
                fa.stunden += ku.stunden;
            }
        }

        // kurs_raum Referenzen zuordnen
        try (ResultSet rs = executeQuery("SELECT kurs_id, raum_id FROM kurs_raum;")) {
            for (int row = 0; rs.next(); row++) {
                D_Kurs ku = map_ku.get_obj_with_id(rs.getInt(1));
                D_Raum ra = map_ra.get_obj_with_id(rs.getInt(2));
                ku.ra_list.add(ra);
                ra.mapKu.put(ku.id, ku);
            }
        }

        // kurs_kopplung Referenzen zuordnen
        try (ResultSet rs = executeQuery("SELECT kurs_id, kopplung_id FROM kurs_kopplung;")) {
            for (int row = 0; rs.next(); row++) {
                D_Kurs ku = map_ku.get_obj_with_id(rs.getInt(1));
                D_Kopplung ko = map_ko.get_obj_with_id(rs.getInt(2));
                ku.ko_list.add(ko);
                ko.mapKu.put(ku.id, ku);
            }
        }

        // kurs_klasse Referenzen zuordnen (nach kurs_kopplung!!!)
        try (ResultSet rs = executeQuery("SELECT kurs_id, klasse_id FROM kurs_klasse ORDER BY klasse_id;")) {
            for (int row = 0; rs.next(); row++) {
                D_Kurs ku = map_ku.get_obj_with_id(rs.getInt(1));
                D_Klasse kl = map_kl.get_obj_with_id(rs.getInt(2));
                ku.kl_list.add(kl);
                kl.mapKu.put(ku.id, ku);
                if (ku.ko_list.isEmpty()) {
                    kl.stunden += ku.stunden;
                } else {
                    D_Kopplung ko = ku.ko_list.getFirst();
                    if (!kl.list_ko.contains(ko)) {
                        kl.list_ko.addLast(ko);
                        kl.stunden += ko.stunden;
                    }
                }
            }
        }

        // Berechnungen aus DB erzeugen
        try (ResultSet rs = executeQuery("SELECT deaktiviere_le_freie_tage, deaktiviere_le_erst_ab_stunde, deaktiviere_le_nur_bis_stunde"
                + ",deaktiviere_le_springstunden_tag, deaktiviere_le_springstunden_woche, deaktiviere_le_tagesstundenpraesenz, deaktiviere_le_min_stunden_pro_tag"
                + ", deaktiviere_le_max_stunden_in_folge, deaktiviere_le_plan_spalte_min, deaktiviere_le_plan_spalte_max, deaktiviere_le_plan_zeile_min"
                + ", deaktiviere_le_plan_zeile_max, deaktiviere_le_plan_zeile_min_ab, deaktiviere_le_plan_zeile_max_ab, deaktiviere_le_plan_zellen FROM berechnungen WHERE id=1;")) {
            for (int row = 0; rs.next(); row++) {
                d_berechnungen.deaktiviere_le_freie_tage = rs.getInt(1) == 1;
                d_berechnungen.deaktiviere_le_erst_ab_stunde = rs.getInt(2) == 1;
                d_berechnungen.deaktiviere_le_nur_bis_stunde = rs.getInt(3) == 1;
                d_berechnungen.deaktiviere_le_springstunden_tag = rs.getInt(4) == 1;
                d_berechnungen.deaktiviere_le_springstunden_woche = rs.getInt(5) == 1;
                d_berechnungen.deaktiviere_le_tagesstundenpraesenz = rs.getInt(6) == 1;
                d_berechnungen.deaktiviere_le_min_stunden_pro_tag = rs.getInt(7) == 1;
                d_berechnungen.deaktiviere_le_max_stunden_in_folge = rs.getInt(8) == 1;
                d_berechnungen.deaktiviere_le_plan_spalte_min = rs.getInt(9) == 1;
                d_berechnungen.deaktiviere_le_plan_spalte_max = rs.getInt(10) == 1;
                d_berechnungen.deaktiviere_le_plan_zeile_min = rs.getInt(11) == 1;
                d_berechnungen.deaktiviere_le_plan_zeile_max = rs.getInt(12) == 1;
                d_berechnungen.deaktiviere_le_plan_zeile_min_ab = rs.getInt(13) == 1;
                d_berechnungen.deaktiviere_le_plan_zeile_max_ab = rs.getInt(14) == 1;
                d_berechnungen.deaktiviere_le_plan_zellen = rs.getInt(15) == 1;
            }
        }

        // Informieren 
        gui.gui_tree_structure_changed_pre(select_same);
        TreeModelEvent tme = new TreeModelEvent(this, tr_get_node_schule());
        for (TreeModelListener tml : listener) {
            tml.treeStructureChanged(tme);
        }
        gui.gui_tree_structure_changed_post(select_same);
    }

    // (8)
    void db_sc_clear() throws Exception {
        // Deletion
        for (String tabelle : SAVE_TABLE_ORDER) {
            db_executeUpdate("DELETE FROM " + tabelle + ";");
        }

        // Singletons
        for (String tabelle : table_with_single_line) {
            db_executeUpdate("INSERT INTO " + tabelle + " (id) VALUES (1);");
        }

        db_sc_reload(false);
    }

    @Override
    public String toString() {
        return "Wurzel";
    }

    @Override
    public Object getRoot() {
        return this;
    }

    @Override
    public void addTreeModelListener(TreeModelListener l) {
        listener.addLast(l);
    }

    @Override
    public void removeTreeModelListener(TreeModelListener l) {
        listener.remove(l);
    }

    @Override
    public void valueForPathChanged(TreePath path, Object newValue) {

    }

    // ### DB CONNECTION ###
    void db_connect_and_create() throws Exception {
        con = DriverManager.getConnection("jdbc:h2:mem:");
        Statement stm = con.createStatement();

        stm.execute("CREATE TABLE schule("
                + "id INT PRIMARY KEY AUTO_INCREMENT NOT NULL"
                + ", max_t INT NOT NULL DEFAULT 5"
                + ", max_s INT NOT NULL DEFAULT 6"
                + ", max_spring_pro_tag INT NOT NULL DEFAULT 4"
                + ", max_spring_pro_woche INT NOT NULL DEFAULT 8"
                + ", max_praesenz_pro_tag INT NOT NULL DEFAULT 9"
                + ", plan VARCHAR NOT NULL DEFAULT ''"
                + ", keine_einzelstunde_in_1 INT NOT NULL DEFAULT 0"
                + ", keine_einzelstunde_in_2 INT NOT NULL DEFAULT 0"
                + ", keine_einzelstunde_in_3 INT NOT NULL DEFAULT 0"
                + ", keine_einzelstunde_in_4 INT NOT NULL DEFAULT 0"
                + ", keine_einzelstunde_in_5 INT NOT NULL DEFAULT 0"
                + ", keine_einzelstunde_in_6 INT NOT NULL DEFAULT 0"
                + ", planV VARCHAR NOT NULL DEFAULT ''"
                + ", keinen_uebergang_1_2 INT NOT NULL DEFAULT 0"
                + ", keinen_uebergang_2_3 INT NOT NULL DEFAULT 0"
                + ", keinen_uebergang_3_4 INT NOT NULL DEFAULT 0"
                + ", keinen_uebergang_4_5 INT NOT NULL DEFAULT 0"
                + ", keinen_uebergang_5_6 INT NOT NULL DEFAULT 0"
                + ", keinen_uebergang_6_7 INT NOT NULL DEFAULT 0"
                + ", keinen_uebergang_7_8 INT NOT NULL DEFAULT 0"
                + ", planO VARCHAR NOT NULL DEFAULT ''"
                + ", min_stunden_le INT NOT NULL DEFAULT 1"
                + ", min_aufeinanderfolgend INT NOT NULL DEFAULT 0"
                + ", max_stunden_folge_le INT NOT NULL DEFAULT 8"
                + ");");

        stm.execute("CREATE TABLE berechnung_schule("
                + "id INT PRIMARY KEY AUTO_INCREMENT NOT NULL"
                + ", status INT NOT NULL"
                + ", millis BIGINT NOT NULL"
                + ", max_t INT NOT NULL"
                + ", max_s INT NOT NULL"
                + ");");

        stm.execute("CREATE TABLE lehrkraefte("
                + "id INT PRIMARY KEY AUTO_INCREMENT NOT NULL"
                + ", plan VARCHAR NOT NULL DEFAULT ''"
                + ");");

        stm.execute("CREATE TABLE lehrkraft("
                + "id INT PRIMARY KEY AUTO_INCREMENT NOT NULL"
                + ", kuerzel VARCHAR(8) NOT NULL DEFAULT '?'"
                + ", anzeigename VARCHAR(64) NOT NULL DEFAULT ''"
                + ", bemerkung VARCHAR(240) NOT NULL DEFAULT ''"
                + ", darf_vertreten INT NOT NULL DEFAULT 1"
                + ", freie_tage INT NOT NULL DEFAULT 0"
                + ", mind_x_ab_stunde INT NOT NULL DEFAULT 0"
                + ", mind_y_ab_stunde INT NOT NULL DEFAULT 0"
                + ", mind_x_bis_stunde INT NOT NULL DEFAULT 0"
                + ", mind_y_bis_stunde INT NOT NULL DEFAULT 0"
                + ", spring_pro_tag_check INT NOT NULL DEFAULT 1"
                + ", spring_pro_tag_wert INT NOT NULL DEFAULT 4"
                + ", spring_pro_woche_check INT NOT NULL DEFAULT 1"
                + ", spring_pro_woche_wert INT NOT NULL DEFAULT 8"
                + ", tagesstunden_check INT NOT NULL DEFAULT 1"
                + ", tagesstunden_wert INT NOT NULL DEFAULT 9"
                + ", plan VARCHAR NOT NULL DEFAULT ''"
                + ", min_stunden_check INT NOT NULL DEFAULT 1"
                + ", min_stunden_wert INT NOT NULL DEFAULT 1"
                + ", max_stunden_folge_check INT NOT NULL DEFAULT 1"
                + ", max_stunden_folge_wert INT NOT NULL DEFAULT 8"
                + ");");

        stm.execute("CREATE TABLE berechnung_lehrkraft("
                + "be_id INT NOT NULL REFERENCES berechnung_schule(id) ON DELETE CASCADE"
                + ", le_id INT NOT NULL"
                + ", kuerzel VARCHAR(8) NOT NULL"
                + ", anzeigename VARCHAR(64) NOT NULL"
                + ", bemerkung VARCHAR(240) NOT NULL"
                + ", darf_vertreten INT NOT NULL DEFAULT 1"
                + ", PRIMARY KEY(be_id, le_id)"
                + ");");

        stm.execute("CREATE TABLE klassen("
                + "id INT PRIMARY KEY AUTO_INCREMENT NOT NULL"
                + ", plan VARCHAR NOT NULL DEFAULT ''"
                + ");");

        stm.execute("CREATE TABLE klasse("
                + "id INT PRIMARY KEY AUTO_INCREMENT NOT NULL"
                + ", kuerzel VARCHAR(8) NOT NULL DEFAULT '?'"
                + ", plan VARCHAR NOT NULL DEFAULT ''"
                + ");");
        stm.execute("CREATE TABLE berechnung_klasse("
                + "be_id INT NOT NULL REFERENCES berechnung_schule(id) ON DELETE CASCADE"
                + ", kl_id INT NOT NULL"
                + ", kuerzel VARCHAR(8) NOT NULL"
                + ", PRIMARY KEY(be_id, kl_id)"
                + ");");

        stm.execute("CREATE TABLE fach("
                + "id INT PRIMARY KEY AUTO_INCREMENT NOT NULL"
                + ", kuerzel VARCHAR(8) NOT NULL DEFAULT '?'"
                + ", anzeigename VARCHAR(64) NOT NULL DEFAULT ''"
                + ");");
        stm.execute("CREATE TABLE berechnung_fach("
                + "be_id INT NOT NULL REFERENCES berechnung_schule(id) ON DELETE CASCADE"
                + ", fa_id INT NOT NULL"
                + ", kuerzel VARCHAR(8) NOT NULL"
                + ", anzeigename VARCHAR(64) NOT NULL"
                + ", PRIMARY KEY(be_id, fa_id)"
                + ");");

        stm.execute("CREATE TABLE raeume("
                + "id INT PRIMARY KEY AUTO_INCREMENT NOT NULL"
                + ", plan VARCHAR NOT NULL DEFAULT ''"
                + ");");
        stm.execute("CREATE TABLE raum("
                + "id INT PRIMARY KEY AUTO_INCREMENT NOT NULL"
                + ", kuerzel VARCHAR(8) NOT NULL DEFAULT '?'"
                + ", plan VARCHAR NOT NULL DEFAULT ''"
                + ");");
        stm.execute("CREATE TABLE berechnung_raum("
                + "be_id INT NOT NULL REFERENCES berechnung_schule(id) ON DELETE CASCADE"
                + ", ra_id INT NOT NULL"
                + ", kuerzel VARCHAR(8) NOT NULL"
                + ", PRIMARY KEY(be_id, ra_id)"
                + ");");

        stm.execute("CREATE TABLE kopplungen("
                + "id INT PRIMARY KEY AUTO_INCREMENT NOT NULL"
                + ", plan VARCHAR NOT NULL DEFAULT ''"
                + ");");
        stm.execute("CREATE TABLE kopplung("
                + "id INT PRIMARY KEY AUTO_INCREMENT NOT NULL"
                + ", kuerzel VARCHAR(8) NOT NULL DEFAULT '?'"
                + ", anzeigename VARCHAR(64) NOT NULL DEFAULT ''"
                + ", plan VARCHAR NOT NULL DEFAULT ''"
                + ");");
        stm.execute("CREATE TABLE berechnung_kopplung("
                + "be_id INT NOT NULL REFERENCES berechnung_schule(id) ON DELETE CASCADE"
                + ", ko_id INT NOT NULL"
                + ", kuerzel VARCHAR(8) NOT NULL"
                + ", anzeigename VARCHAR(64) NOT NULL"
                + ", PRIMARY KEY(be_id, ko_id)"
                + ");");

        stm.execute("CREATE TABLE kurse("
                + "id INT PRIMARY KEY AUTO_INCREMENT NOT NULL"
                + ", plan VARCHAR NOT NULL DEFAULT ''"
                + ", sortierung INT NOT NULL DEFAULT 0"
                + ");");

        stm.execute("CREATE TABLE kurs("
                + "id INT PRIMARY KEY AUTO_INCREMENT NOT NULL"
                + ", ergaenzung VARCHAR(8) NOT NULL DEFAULT ''"
                + ", vertretungsreserven INT NOT NULL DEFAULT 0"
                + ", plan VARCHAR NOT NULL DEFAULT ''"
                + ", nicht_aufeinanderfolgend INT NOT NULL DEFAULT 0"
                + ");");
        stm.execute("CREATE TABLE berechnung_kurs("
                + "be_id INT NOT NULL REFERENCES berechnung_schule(id) ON DELETE CASCADE"
                + ", ku_id INT NOT NULL"
                + ", ergaenzung VARCHAR(8) NOT NULL"
                + ", vertretungsreserven INT NOT NULL DEFAULT 0"
                + ", PRIMARY KEY(be_id, ku_id)"
                + ");");

        stm.execute("CREATE TABLE kurs_kurs("
                + "kurs_id_1 INT NOT NULL REFERENCES kurs(id) ON DELETE CASCADE"
                + ", kurs_id_2 INT NOT NULL REFERENCES kurs(id) ON DELETE CASCADE"
                + ", PRIMARY KEY(kurs_id_1, kurs_id_2)"
                + ");");

        stm.execute("CREATE TABLE kurs_lehrkraft("
                + "kurs_id INT NOT NULL REFERENCES kurs(id) ON DELETE CASCADE"
                + ", lehrkraft_id INT NOT NULL REFERENCES lehrkraft(id) ON DELETE CASCADE"
                + ", PRIMARY KEY(kurs_id, lehrkraft_id)"
                + ");");
        stm.execute("CREATE TABLE berechnung_kurs_lehrkraft("
                + "be_id INT NOT NULL REFERENCES berechnung_schule(id) ON DELETE CASCADE"
                + ", ku_id INT NOT NULL"
                + ", le_id INT NOT NULL"
                + ", PRIMARY KEY(be_id, ku_id, le_id)"
                + ");");

        stm.execute("CREATE TABLE kurs_hosp("
                + "kurs_id INT NOT NULL REFERENCES kurs(id) ON DELETE CASCADE"
                + ", lehrkraft_id INT NOT NULL REFERENCES lehrkraft(id) ON DELETE CASCADE"
                + ", PRIMARY KEY(kurs_id, lehrkraft_id)"
                + ");");
        stm.execute("CREATE TABLE berechnung_kurs_hosp("
                + "be_id INT NOT NULL REFERENCES berechnung_schule(id) ON DELETE CASCADE"
                + ", ku_id INT NOT NULL"
                + ", le_id INT NOT NULL"
                + ", PRIMARY KEY(be_id, ku_id, le_id)"
                + ");");

        stm.execute("CREATE TABLE kurs_klasse("
                + "kurs_id INT NOT NULL REFERENCES kurs(id) ON DELETE CASCADE"
                + ", klasse_id INT NOT NULL REFERENCES klasse(id) ON DELETE CASCADE"
                + ", PRIMARY KEY(kurs_id, klasse_id)"
                + ");");
        stm.execute("CREATE TABLE berechnung_kurs_klasse("
                + "be_id INT NOT NULL REFERENCES berechnung_schule(id) ON DELETE CASCADE"
                + ", ku_id INT NOT NULL"
                + ", kl_id INT NOT NULL"
                + ", PRIMARY KEY(be_id, ku_id, kl_id)"
                + ");");

        stm.execute("CREATE TABLE kurs_fach("
                + "kurs_id INT NOT NULL REFERENCES kurs(id) ON DELETE CASCADE"
                + ", fach_id INT NOT NULL REFERENCES fach(id) ON DELETE CASCADE"
                + ", PRIMARY KEY(kurs_id, fach_id)"
                + ");");
        stm.execute("CREATE TABLE berechnung_kurs_fach("
                + "be_id INT NOT NULL REFERENCES berechnung_schule(id) ON DELETE CASCADE"
                + ", ku_id INT NOT NULL"
                + ", fa_id INT NOT NULL"
                + ", PRIMARY KEY(be_id, ku_id, fa_id)"
                + ");");

        stm.execute("CREATE TABLE kurs_raum("
                + "kurs_id INT NOT NULL REFERENCES kurs(id) ON DELETE CASCADE"
                + ", raum_id INT NOT NULL REFERENCES raum(id) ON DELETE CASCADE"
                + ", PRIMARY KEY(kurs_id, raum_id)"
                + ");");
        stm.execute("CREATE TABLE berechnung_kurs_raum("
                + "be_id INT NOT NULL REFERENCES berechnung_schule(id) ON DELETE CASCADE"
                + ", ku_id INT NOT NULL"
                + ", ra_id INT NOT NULL"
                + ", PRIMARY KEY(be_id, ku_id, ra_id)"
                + ");");

        stm.execute("CREATE TABLE kurs_kopplung("
                + "kurs_id INT NOT NULL REFERENCES kurs(id) ON DELETE CASCADE"
                + ", kopplung_id INT NOT NULL REFERENCES kopplung(id) ON DELETE CASCADE"
                + ", PRIMARY KEY(kurs_id, kopplung_id)"
                + ");");
        stm.execute("CREATE TABLE berechnung_kurs_kopplung("
                + "be_id INT NOT NULL REFERENCES berechnung_schule(id) ON DELETE CASCADE"
                + ", ku_id INT NOT NULL"
                + ", ko_id INT NOT NULL"
                + ", PRIMARY KEY(be_id, ku_id, ko_id)"
                + ");");

        stm.execute("CREATE TABLE stunde("
                + "id INT PRIMARY KEY AUTO_INCREMENT NOT NULL"
                + ", kurs_id INT NOT NULL REFERENCES kurs(id) ON DELETE CASCADE"
                + ", wochenstunden INT NOT NULL DEFAULT " + 1
                + ", plan VARCHAR NOT NULL DEFAULT ''"
                + ", typ INT NOT NULL DEFAULT " + Config.STUNDEN_TYP_N
                + ", lage_kl INT NOT NULL DEFAULT " + Config.STUNDEN_LAGE_EGAL
                + ", lage_le INT NOT NULL DEFAULT " + Config.STUNDEN_LAGE_EGAL
                + ");");

        stm.execute("CREATE TABLE berechnung_stunde("
                + "be_id INT NOT NULL REFERENCES berechnung_schule(id) ON DELETE CASCADE"
                + ", ku_id INT NOT NULL"
                + ", w INT NOT NULL" // woche
                + ", t INT NOT NULL" // tag
                + ", s INT NOT NULL" // stunde
                + ", PRIMARY KEY(be_id, ku_id, w, t, s)"
                + ");");

        stm.execute("CREATE TABLE stunde_ko("
                + "id INT PRIMARY KEY AUTO_INCREMENT NOT NULL"
                + ", kopplung_id INT NOT NULL REFERENCES kopplung(id) ON DELETE CASCADE"
                + ", wochenstunden INT NOT NULL DEFAULT " + 1
                + ", plan VARCHAR NOT NULL DEFAULT ''"
                + ", typ INT NOT NULL DEFAULT " + Config.STUNDEN_TYP_N
                + ", lage_kl INT NOT NULL DEFAULT " + Config.STUNDEN_LAGE_EGAL
                + ", lage_le INT NOT NULL DEFAULT " + Config.STUNDEN_LAGE_EGAL
                + ");");
        stm.execute("CREATE TABLE berechnung_stunde_ko("
                + "be_id INT NOT NULL REFERENCES berechnung_schule(id) ON DELETE CASCADE"
                + ", ko_id INT NOT NULL"
                + ", w INT NOT NULL" // woche
                + ", t INT NOT NULL" // tag
                + ", s INT NOT NULL" // stunde
                + ", PRIMARY KEY(be_id, ko_id, w, t, s)"
                + ");");

        stm.execute("CREATE TABLE berechnungen("
                + "id INT PRIMARY KEY AUTO_INCREMENT NOT NULL"
                + ", deaktiviere_le_freie_tage INT NOT NULL DEFAULT 0"
                + ", deaktiviere_le_erst_ab_stunde INT NOT NULL DEFAULT 0"
                + ", deaktiviere_le_nur_bis_stunde INT NOT NULL DEFAULT 0"
                + ", deaktiviere_le_springstunden_tag INT NOT NULL DEFAULT 0"
                + ", deaktiviere_le_springstunden_woche INT NOT NULL DEFAULT 0"
                + ", deaktiviere_le_tagesstundenpraesenz INT NOT NULL DEFAULT 0"
                + ", deaktiviere_le_min_stunden_pro_tag INT NOT NULL DEFAULT 0"
                + ", deaktiviere_le_max_stunden_in_folge INT NOT NULL DEFAULT 0"
                + ", deaktiviere_le_plan_spalte_min INT NOT NULL DEFAULT 0"
                + ", deaktiviere_le_plan_spalte_max INT NOT NULL DEFAULT 0"
                + ", deaktiviere_le_plan_zeile_min INT NOT NULL DEFAULT 0"
                + ", deaktiviere_le_plan_zeile_max INT NOT NULL DEFAULT 0"
                + ", deaktiviere_le_plan_zeile_min_ab INT NOT NULL DEFAULT 0"
                + ", deaktiviere_le_plan_zeile_max_ab INT NOT NULL DEFAULT 0"
                + ", deaktiviere_le_plan_zellen INT NOT NULL DEFAULT 0"
                + ");");

        db_sc_clear(); // for loading default values
    }

    int db_executeUpdate_key(String sql) throws Exception {
        Statement stm = con.createStatement();
        stm.executeUpdate(sql, Statement.RETURN_GENERATED_KEYS);
        try (ResultSet rs = stm.getGeneratedKeys()) {
            while (rs.next()) {
                return rs.getInt(1);
            }
        }
        throw new Exception("Primärschlüssel-Fehler: " + sql);
    }

    void db_executeUpdate(String sql) throws Exception {
        Statement stm = con.createStatement();
        stm.executeUpdate(sql);
    }

    ResultSet executeQuery(String sql) throws SQLException {
        Statement stm = con.createStatement();
        return stm.executeQuery(sql);
    }

    //
    // ########## DATENBANK ##########
    //
    // ########## DB SC ##########
    void db_sc_change_wochentage(int id, int max_t) throws Exception {
        db_executeUpdate("UPDATE schule SET max_t=" + max_t + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_sc_change_stunden_pro_tag(int id, int max_s) throws Exception {
        db_executeUpdate("UPDATE schule SET max_s=" + max_s + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_sc_change_max_spring_pro_tag(int id, int p_wert) throws Exception {
        db_executeUpdate("UPDATE schule SET max_spring_pro_tag=" + p_wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_sc_change_max_spring_pro_woche(int id, int p_wert) throws Exception {
        db_executeUpdate("UPDATE schule SET max_spring_pro_woche=" + p_wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_sc_change_max_tagespraesenz(int id, int p_wert) throws Exception {
        db_executeUpdate("UPDATE schule SET max_praesenz_pro_tag=" + p_wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_sc_change_keine_einzelstunde_in_1(int id, boolean wert) throws Exception {
        db_executeUpdate("UPDATE schule SET keine_einzelstunde_in_1=" + (wert ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_sc_change_keine_einzelstunde_in_2(int id, boolean wert) throws Exception {
        db_executeUpdate("UPDATE schule SET keine_einzelstunde_in_2=" + (wert ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_sc_change_keine_einzelstunde_in_3(int id, boolean wert) throws Exception {
        db_executeUpdate("UPDATE schule SET keine_einzelstunde_in_3=" + (wert ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_sc_change_keine_einzelstunde_in_4(int id, boolean wert) throws Exception {
        db_executeUpdate("UPDATE schule SET keine_einzelstunde_in_4=" + (wert ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_sc_change_keine_einzelstunde_in_5(int id, boolean wert) throws Exception {
        db_executeUpdate("UPDATE schule SET keine_einzelstunde_in_5=" + (wert ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_sc_change_keine_einzelstunde_in_6(int id, boolean wert) throws Exception {
        db_executeUpdate("UPDATE schule SET keine_einzelstunde_in_6=" + (wert ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_sc_change_keinen_uebergang_1_2(int id, boolean wert) throws Exception {
        db_executeUpdate("UPDATE schule SET keinen_uebergang_1_2=" + (wert ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_sc_change_keinen_uebergang_2_3(int id, boolean wert) throws Exception {
        db_executeUpdate("UPDATE schule SET keinen_uebergang_2_3=" + (wert ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_sc_change_keinen_uebergang_3_4(int id, boolean wert) throws Exception {
        db_executeUpdate("UPDATE schule SET keinen_uebergang_3_4=" + (wert ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_sc_change_keinen_uebergang_4_5(int id, boolean wert) throws Exception {
        db_executeUpdate("UPDATE schule SET keinen_uebergang_4_5=" + (wert ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_sc_change_keinen_uebergang_5_6(int id, boolean wert) throws Exception {
        db_executeUpdate("UPDATE schule SET keinen_uebergang_5_6=" + (wert ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_sc_change_keinen_uebergang_6_7(int id, boolean wert) throws Exception {
        db_executeUpdate("UPDATE schule SET keinen_uebergang_6_7=" + (wert ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_sc_change_keinen_uebergang_7_8(int id, boolean wert) throws Exception {
        db_executeUpdate("UPDATE schule SET keinen_uebergang_7_8=" + (wert ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_sc_le_change_min_stunden_pro_tag(int id, int wert) throws Exception {
        db_executeUpdate("UPDATE schule SET min_stunden_le =" + wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_sc_le_change_max_stunden_in_folge(int id, int wert) throws Exception {
        db_executeUpdate("UPDATE schule SET max_stunden_folge_le =" + wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_sc_change_min_aufeinanderfolgend(int id, int wert) throws Exception {
        wert = Math.min(wert, d_schule.min_aufeinanderfolgend_von);
        db_executeUpdate("UPDATE schule SET min_aufeinanderfolgend =" + wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_sc_plaene_loeschen() throws Exception {
        String[] tabellen_mit_plan = new String[]{
            "schule",
            "lehrkraefte",
            "lehrkraft",
            "klassen",
            "klasse",
            "raeume",
            "raum",
            "kopplungen",
            "kopplung",
            "kurse",
            "kurs",
            "stunde",
            "stunde_ko"
        };
        for (String t : tabellen_mit_plan) {
            db_executeUpdate("UPDATE " + t + " SET plan='';");
        }
        db_sc_reload(false);
    }

    // ########## DB LE ##########
    int db_le_erzeugen(String p_kuerzel, String p_anzeigename, String p_bemerkung) throws Exception {
        int id = db_executeUpdate_key("INSERT INTO lehrkraft (id) VALUES (NULL);");
        db_sc_reload(false);
        return id;
    }

    void db_le_alle_loeschen() throws Exception {
        db_executeUpdate("DELETE FROM lehrkraft;");
        db_sc_reload(true);
    }

    void db_le_loesche(int id) throws Exception {
        db_executeUpdate("DELETE FROM lehrkraft WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_le_change_kuerzel(int id, String text) throws Exception {
        db_executeUpdate("UPDATE lehrkraft SET kuerzel='" + text + "' WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_le_change_anzeigename(int id, String text) throws Exception {
        db_executeUpdate("UPDATE lehrkraft SET anzeigename='" + text + "' WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_le_change_bemerkung(int id, String text) throws Exception {
        db_executeUpdate("UPDATE lehrkraft SET bemerkung='" + text + "' WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_le_change_darf_vertreten(int id, boolean selected) throws Exception {
        int wert = selected ? 1 : 0;
        db_executeUpdate("UPDATE lehrkraft SET darf_vertreten=" + wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_le_change_freie_tag(int id, int wert) throws Exception {
        db_executeUpdate("UPDATE lehrkraft SET freie_tage=" + wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_le_change_ab_stunde_x(int id, int wert) throws Exception {
        db_executeUpdate("UPDATE lehrkraft SET mind_x_ab_stunde=" + wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_le_change_ab_stunde_y(int id, int wert) throws Exception {
        db_executeUpdate("UPDATE lehrkraft SET mind_y_ab_stunde=" + wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_le_change_bis_stunde_x(int id, int wert) throws Exception {
        db_executeUpdate("UPDATE lehrkraft SET mind_x_bis_stunde=" + wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_le_change_bis_stunde_y(int id, int wert) throws Exception {
        db_executeUpdate("UPDATE lehrkraft SET mind_y_bis_stunde=" + wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_le_change_spring_pro_tag_check(int id, boolean b) throws Exception {
        int wert = b ? 1 : 0;
        db_executeUpdate("UPDATE lehrkraft SET spring_pro_tag_check=" + wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_le_change_spring_pro_tag_wert(int id, int wert) throws Exception {
        db_executeUpdate("UPDATE lehrkraft SET spring_pro_tag_wert=" + wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_le_change_spring_pro_woche_check(int id, boolean b) throws Exception {
        int wert = b ? 1 : 0;
        db_executeUpdate("UPDATE lehrkraft SET spring_pro_woche_check=" + wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_le_change_spring_pro_woche_wert(int id, int wert) throws Exception {
        db_executeUpdate("UPDATE lehrkraft SET spring_pro_woche_wert=" + wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_le_change_tagesstunden_check(int id, boolean b) throws Exception {
        int wert = b ? 1 : 0;
        db_executeUpdate("UPDATE lehrkraft SET tagesstunden_check=" + wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_le_change_tagesstunden_wert(int id, int wert) throws Exception {
        db_executeUpdate("UPDATE lehrkraft SET tagesstunden_wert=" + wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_le_change_min_stunden_check(int id, boolean b) throws Exception {
        int wert = b ? 1 : 0;
        db_executeUpdate("UPDATE lehrkraft SET min_stunden_check=" + wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_le_min_stunden_wert(int id, int wert) throws Exception {
        db_executeUpdate("UPDATE lehrkraft SET min_stunden_wert=" + wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_le_change_max_stunden_folge_check(int id, boolean b) throws Exception {
        int wert = b ? 1 : 0;
        db_executeUpdate("UPDATE lehrkraft SET max_stunden_folge_check=" + wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_le_max_stunden_folge_wert(int id, int wert) throws Exception {
        db_executeUpdate("UPDATE lehrkraft SET max_stunden_folge_wert=" + wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    // ########## DB KL ##########
    int db_kl_erzeugen() throws Exception {
        int id = db_executeUpdate_key("INSERT INTO klasse (id) VALUES (NULL);");
        db_sc_reload(false);
        return id;
    }

    void db_kl_erzeuge_standardklassen(String[] p_prefix, String[] p_suffix) throws Exception {
        db_executeUpdate("DELETE FROM klasse;");
        for (int i = 0; i < p_prefix.length; i++) {
            for (int j = 0; j < p_suffix.length; j++) {
                db_executeUpdate("INSERT INTO klasse (id, kuerzel) VALUES (NULL, '" + p_prefix[i] + p_suffix[j] + "');");
            }
        }
        db_sc_reload(false);
    }

    void db_kl_alle_loeschen() throws Exception {
        db_executeUpdate("DELETE FROM klasse;");
        db_sc_reload(false);
    }

    void db_kl_loesche(int id) throws Exception {
        db_executeUpdate("DELETE FROM klasse WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_kl_change_kuerzel(int id, String text) throws Exception {
        db_executeUpdate("UPDATE klasse SET kuerzel='" + text + "' WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    // ########## DB FA ##########
    int db_fa_erzeugen() throws Exception {
        int id = db_executeUpdate_key("INSERT INTO fach (id) VALUES (NULL);");
        db_sc_reload(false);
        return id;
    }

    void db_fa_erzeuge_standard_grundschule() throws Exception {
        db_executeUpdate("DELETE FROM fach;");

        String[][] temp = new String[][]{
            {"D", "Deutsch"},
            {"Ma", "Mathematik"},
            {"En", "Englisch"},
            {"Sa", "Sachunterricht"},
            {"Ku", "Kunst"},
            {"Mu", "Musik"},
            {"Sp", "Sport"},
            {"Re", "Religion"}};

        for (int i = 0; i < temp.length; i++) {
            db_executeUpdate("INSERT INTO fach (id, kuerzel, anzeigename) VALUES(NULL, '" + temp[i][0] + "', '" + temp[i][1] + "');");
        }

        db_sc_reload(false);
    }

    void db_fa_alle_loeschen() throws Exception {
        db_executeUpdate("DELETE FROM fach;");
        db_sc_reload(false);
    }

    void db_fa_loesche(int id) throws Exception {
        db_executeUpdate("DELETE FROM fach WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_fa_change_kuerzel(int id, String text) throws Exception {
        db_executeUpdate("UPDATE fach SET kuerzel='" + text + "' WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_fa_change_anzeigename(int id, String text) throws Exception {
        db_executeUpdate("UPDATE fach SET anzeigename='" + text + "' WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    // ########## DB RA ##########
    int db_ra_erzeugen(String p_kuerzel) throws Exception {
        int id = db_executeUpdate_key("INSERT INTO raum (id) VALUES (NULL);");
        db_sc_reload(false);
        return id;
    }

    void db_ra_alle_loeschen() throws Exception {
        db_executeUpdate("DELETE FROM raum;");
        db_sc_reload(false);
    }

    void db_ra_loesche(int id) throws Exception {
        db_executeUpdate("DELETE FROM raum WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_ra_change_kuerzel(int id, String text) throws Exception {
        db_executeUpdate("UPDATE raum SET kuerzel='" + text + "' WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    // ########## DB KO ##########
    int db_ko_erzeugen() throws Exception {
        int id = db_executeUpdate_key("INSERT INTO kopplung (id) VALUES (NULL);");
        db_sc_reload(false);
        return id;
    }

    void db_ko_alle_loeschen() throws Exception {
        db_executeUpdate("DELETE FROM kopplung;");
        db_sc_reload(false);
    }

    void db_ko_loesche(int id) throws Exception {
        db_executeUpdate("DELETE FROM kopplung WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_ko_change_kuerzel(int id, String text) throws Exception {
        db_executeUpdate("UPDATE kopplung SET kuerzel='" + text + "' WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_ko_change_anzeigename(int id, String text) throws Exception {
        db_executeUpdate("UPDATE kopplung SET anzeigename='" + text + "' WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    // ########## DB KU ##########
    int db_ku_erzeugen() throws Exception {
        int id = db_executeUpdate_key("INSERT INTO kurs (id) VALUES (NULL);");
        db_sc_reload(false);
        return id;
    }

    void db_ku_alle_loeschen() throws Exception {
        db_executeUpdate("DELETE FROM kurs;");
        db_sc_reload(false);
    }

    void db_ku_loesche(int id) throws Exception {
        db_executeUpdate("DELETE FROM kurs WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_ku_change_ergaenzung(int id, String text) throws Exception {
        db_executeUpdate("UPDATE kurs SET ergaenzung='" + text + "' WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_ku_change_vertretungsreserven(int id, int vertretungsreserven) throws Exception {
        db_executeUpdate("UPDATE kurs SET vertretungsreserven=" + vertretungsreserven + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_ku_change_nicht_aufeinanderfolgend(int id, int wert) throws Exception {
        db_executeUpdate("UPDATE kurs SET nicht_aufeinanderfolgend=" + wert + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    // ########## DB KUA ##########
    void db_kua_sortiere_nach(int id, int sortierung) throws Exception {
        db_executeUpdate("UPDATE kurse SET sortierung=" + sortierung + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    // ########## DB ST ##########
    int db_st_erzeugen(int p_ku_id) throws Exception {
        int id = db_executeUpdate_key("INSERT INTO stunde (id, kurs_id) VALUES(NULL, " + p_ku_id + ");");
        db_sc_reload(false);
        return id;
    }

    void db_st_loesche(int id) throws Exception {
        db_executeUpdate("DELETE FROM stunde WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_st_change_wochenstunden(int id, int wochenstunden) throws Exception {
        db_executeUpdate("UPDATE stunde SET wochenstunden=" + wochenstunden + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_st_change_typ(int id, int typ) throws Exception {
        db_executeUpdate("UPDATE stunde SET typ=" + typ + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_st_change_lage_kl(int id, int lage) throws Exception {
        db_executeUpdate("UPDATE stunde SET lage_kl=" + lage + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_st_change_lage_le(int id, int lage) throws Exception {
        db_executeUpdate("UPDATE stunde SET lage_le=" + lage + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    // ########## DB ST KO ##########
    int db_st_ko_erzeugen(int ko_id) throws Exception {
        int id = db_executeUpdate_key("INSERT INTO stunde_ko (id, kopplung_id) VALUES(NULL, " + ko_id + ");");
        db_sc_reload(false);
        return id;
    }

    void db_st_ko_loesche(int id) throws Exception {
        db_executeUpdate("DELETE FROM stunde_ko WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_st_ko_change_wochenstunden(int id, int wochenstunden) throws Exception {
        db_executeUpdate("UPDATE stunde_ko SET wochenstunden=" + wochenstunden + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_st_ko_change_typ(int id, int typ) throws Exception {
        db_executeUpdate("UPDATE stunde_ko SET typ=" + typ + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_st_ko_change_lage_kl(int id, int lage) throws Exception {
        db_executeUpdate("UPDATE stunde_ko SET lage_kl=" + lage + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_st_ko_change_lage_le(int id, int lage) throws Exception {
        db_executeUpdate("UPDATE stunde_ko SET lage_le=" + lage + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    // ########## DB BEA ##########
    void db_bea_change_deaktiviere_le_freie_tage(int id, boolean b) throws Exception {
        db_executeUpdate("UPDATE berechnungen SET deaktiviere_le_freie_tage=" + (b ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_bea_change_deaktiviere_le_erst_ab_stunde(int id, boolean b) throws Exception {
        db_executeUpdate("UPDATE berechnungen SET deaktiviere_le_erst_ab_stunde=" + (b ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_bea_change_deaktiviere_le_nur_bis_stunde(int id, boolean b) throws Exception {
        db_executeUpdate("UPDATE berechnungen SET deaktiviere_le_nur_bis_stunde=" + (b ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_bea_change_deaktiviere_le_springstunden_tag(int id, boolean b) throws Exception {
        db_executeUpdate("UPDATE berechnungen SET deaktiviere_le_springstunden_tag=" + (b ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_bea_change_deaktiviere_le_springstunden_woche(int id, boolean b) throws Exception {
        db_executeUpdate("UPDATE berechnungen SET deaktiviere_le_springstunden_woche=" + (b ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_bea_change_deaktiviere_le_tagesstundenpraesenz(int id, boolean b) throws Exception {
        db_executeUpdate("UPDATE berechnungen SET deaktiviere_le_tagesstundenpraesenz=" + (b ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_bea_change_deaktiviere_le_min_stunden_pro_tag(int id, boolean b) throws Exception {
        db_executeUpdate("UPDATE berechnungen SET deaktiviere_le_min_stunden_pro_tag=" + (b ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_bea_change_deaktiviere_le_max_stunden_in_folge(int id, boolean b) throws Exception {
        db_executeUpdate("UPDATE berechnungen SET deaktiviere_le_max_stunden_in_folge=" + (b ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_bea_change_deaktiviere_le_plan_spalte_min(int id, boolean b) throws Exception {
        db_executeUpdate("UPDATE berechnungen SET deaktiviere_le_plan_spalte_min=" + (b ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_bea_change_deaktiviere_le_plan_spalte_max(int id, boolean b) throws Exception {
        db_executeUpdate("UPDATE berechnungen SET deaktiviere_le_plan_spalte_max=" + (b ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_bea_change_deaktiviere_le_plan_zeile_min(int id, boolean b) throws Exception {
        db_executeUpdate("UPDATE berechnungen SET deaktiviere_le_plan_zeile_min=" + (b ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_bea_change_deaktiviere_le_plan_zeile_max(int id, boolean b) throws Exception {
        db_executeUpdate("UPDATE berechnungen SET deaktiviere_le_plan_zeile_max=" + (b ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_bea_change_deaktiviere_le_plan_zeile_min_ab(int id, boolean b) throws Exception {
        db_executeUpdate("UPDATE berechnungen SET deaktiviere_le_plan_zeile_min_ab=" + (b ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_bea_change_deaktiviere_le_plan_zeile_max_ab(int id, boolean b) throws Exception {
        db_executeUpdate("UPDATE berechnungen SET deaktiviere_le_plan_zeile_max_ab=" + (b ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    void db_bea_change_deaktiviere_le_plan_zellen(int id, boolean b) throws Exception {
        db_executeUpdate("UPDATE berechnungen SET deaktiviere_le_plan_zellen=" + (b ? 1 : 0) + " WHERE id=" + id + ";");
        db_sc_reload(false);
    }

    // ########## DB BE ##########
    int db_be_erzeugen() throws Exception {
        // Alles kopieren
        D_Schule_SAT sc_sat = new D_Schule_SAT(d_schule, d_lehrkraefte, d_klassen, d_raeume, d_kopplungen, d_kurse, d_berechnungen, map_ku, map_le, map_kl, map_fa, map_ra, map_ko);

        // Berechnung hinzufügen
        int id = db_executeUpdate_key("INSERT INTO berechnung_schule VALUES(NULL, 0, 0, " + sc_sat.MAX_T + ", " + sc_sat.MAX_S + ");");

        // Berechnung starten
        String pfad = Config.get_solver_crypto_pfad(); // TODO CHECK IF EXISTS
        D_Worker worker = new D_Worker(id, sc_sat, pfad);
        map_process.put(id, worker);
        worker.start();

        // reload
        db_sc_reload(false);
        return id;
    }

    void db_be_fertig(int id, int status, long millis, SatWrapper sw, D_Schule_SAT sc_sat) throws Exception {
        // Ergebnis der Berechnung erzeugen
        map_process.remove(id);
        db_executeUpdate("UPDATE berechnung_schule SET status=" + status + ", millis=" + millis + " WHERE id=" + id + ";");

        // ABBRECHEN, falls nicht SAT
        if (status != Config.SOLVER_SAT) {
            db_sc_reload(true);
            //gui.gui_select_path(tr_get_node_berechnung(id), true, false);
            return;
        }

        // Schule_SAT --> DB SOL
        D_Berechnung be = map_be.get_obj_with_id(id);

        // berechnung_lehrkraft (be_id, le_id, kuerzel, anzeigename)
        for (int row = 0; row < sc_sat.myLe.size(); row++) {
            D_Lehrkraft_SAT le_sat = sc_sat.myLe.get_obj_with_row(row);
            int darf_vertreten = le_sat.darf_vertreten ? 1 : 0;
            db_executeUpdate("INSERT INTO berechnung_lehrkraft VALUES(" + be.id + ", " + le_sat.id + ", '" + le_sat.kuerzel + "', '" + le_sat.anzeigename + "', '" + le_sat.bemerkung + "', " + darf_vertreten + ")");
            // le_sat.println_debug(sw);
        }

        // berechnung_klasse (be_id, kl_id, kuerzel)
        for (int row = 0; row < sc_sat.myKl.size(); row++) {
            D_Klasse_SAT kl_sat = sc_sat.myKl.get_obj_with_row(row);
            db_executeUpdate("INSERT INTO berechnung_klasse VALUES(" + be.id + ", " + kl_sat.id + ", '" + kl_sat.kuerzel + "')");
        }

        // berechnung_fach (be_id, fa_id, kuerzel, anzeigename)
        for (int row = 0; row < sc_sat.myFa.size(); row++) {
            D_Fach_SAT fa_sat = sc_sat.myFa.get_obj_with_row(row);
            db_executeUpdate("INSERT INTO berechnung_fach VALUES(" + be.id + ", " + fa_sat.id + ", '" + fa_sat.kuerzel + "', '" + fa_sat.anzeigename + "')");
        }

        // berechnung_raum (be_id, ra_id, kuerzel)
        for (int row = 0; row < sc_sat.myRa.size(); row++) {
            D_Raum_SAT ra_sat = sc_sat.myRa.get_obj_with_row(row);
            db_executeUpdate("INSERT INTO berechnung_raum VALUES(" + be.id + ", " + ra_sat.id + ", '" + ra_sat.kuerzel + "')");
        }

        // berechnung_kopplung (be_id, ko_id, kuerzel, anzeigename)
        for (int row = 0; row < sc_sat.myKo.size(); row++) {
            D_Kopplung_SAT ko_sat = sc_sat.myKo.get_obj_with_row(row);
            db_executeUpdate("INSERT INTO berechnung_kopplung VALUES(" + be.id + ", " + ko_sat.id + ", '" + ko_sat.kuerzel + "', '" + ko_sat.anzeigename + "')");

        }

        // berechnung_kurs ("be_id, ko_id, ergaenzung, vertretungsreserven)
        for (int row = 0; row < sc_sat.myKu.size(); row++) {
            D_Kurs_SAT ku_sat = sc_sat.myKu.get_obj_with_row(row);
            db_executeUpdate("INSERT INTO berechnung_kurs VALUES(" + be.id + ", " + ku_sat.id + ", '" + ku_sat.ergaenzung + "', " + ku_sat.vertretungsreserven + ")");

            // berechnung_stunde (be_id, ko_id, w, t, s)
            for (int w = 0; w < 2; w++) {
                for (int t = 0; t < sc_sat.MAX_T; t++) {
                    for (int s = 0; s < sc_sat.MAX_S; s++) {
                        if (sw.is_true(ku_sat.planWTS[w][t][s])) {
                            db_executeUpdate("INSERT INTO berechnung_stunde VALUES(" + be.id + ", " + ku_sat.id + ", " + w + ", " + t + ", " + s + ")");
                        }
                    }
                }
            }
        }

        // berechnung_kurs_lehrkraft (be_id, ku_id, le_id)
        for (int row = 0; row < sc_sat.myLe.size(); row++) {
            D_Lehrkraft_SAT le_sat = sc_sat.myLe.get_obj_with_row(row);
            for (D_Kurs_SAT ku_sat : le_sat.myKu) {
                db_executeUpdate("INSERT INTO berechnung_kurs_lehrkraft VALUES(" + be.id + ", " + ku_sat.id + ", " + le_sat.id + ")");
            }
        }

        // berechnung_kurs_hosp (be_id, ku_id, le_id)
        for (int row = 0; row < sc_sat.myLe.size(); row++) { // Gehe alle Lehrkräfte durch ...
            D_Lehrkraft_SAT le_sat = sc_sat.myLe.get_obj_with_row(row);
            for (D_Kurs_SAT ku_sat : le_sat.myKuHosp) { // Gehe die Hospitationskurse der Lehrkraft durch
                db_executeUpdate("INSERT INTO berechnung_kurs_hosp VALUES(" + be.id + ", " + ku_sat.id + ", " + le_sat.id + ")");
            }
        }

        // berechnung_kurs_klasse (be_id, ko_id, kl_id)
        for (int row = 0; row < sc_sat.myKl.size(); row++) {
            D_Klasse_SAT kl_sat = sc_sat.myKl.get_obj_with_row(row);
            for (D_Kurs_SAT ku_sat : kl_sat.myKu) {
                db_executeUpdate("INSERT INTO berechnung_kurs_klasse VALUES(" + be.id + ", " + ku_sat.id + ", " + kl_sat.id + ")");
            }
        }

        // berechnung_kurs_fach (be_id, ko_id, fa_id)
        for (int row = 0; row < sc_sat.myFa.size(); row++) {
            D_Fach_SAT fa_sat = sc_sat.myFa.get_obj_with_row(row);
            for (D_Kurs_SAT ku_sat : fa_sat.myKu) {
                db_executeUpdate("INSERT INTO berechnung_kurs_fach VALUES(" + be.id + ", " + ku_sat.id + ", " + fa_sat.id + ")");
            }
        }

        // berechnung_kurs_raum (be_id, ko_id, ra_id)
        for (int row = 0; row < sc_sat.myRa.size(); row++) {
            D_Raum_SAT ra_sat = sc_sat.myRa.get_obj_with_row(row);
            for (D_Kurs_SAT ku_sat : ra_sat.myKu) {
                int varR = ku_sat.planR.get(ra_sat);
                if (sw.is_true(varR)) {
                    db_executeUpdate("INSERT INTO berechnung_kurs_raum VALUES(" + be.id + ", " + ku_sat.id + ", " + ra_sat.id + ")");
                }
            }
        }

        // berechnung_kurs_kopplung (be_id, ko_id, ko_id)
        for (int row = 0; row < sc_sat.myKo.size(); row++) {
            D_Kopplung_SAT ko_sat = sc_sat.myKo.get_obj_with_row(row);
            for (D_Kurs_SAT ku_sat : ko_sat.myKu) {
                db_executeUpdate("INSERT INTO berechnung_kurs_kopplung VALUES(" + be.id + ", " + ku_sat.id + ", " + ko_sat.id + ")");
            }
        }

        // Neuladen der Berechnungen
        db_sc_reload(true);
        //gui.gui_select_path(tr_get_node_berechnung(id), true, false);
    }

    D_Schule_SOL db_be_get_ergebnis(int be_id) throws Exception {
        D_Schule_SOL sc_sol;

        // Berechnung laden
        // TODO berechnung_schule --> max_w?
        try (ResultSet rs = executeQuery("SELECT id, status, millis, max_t, max_s FROM berechnung_schule" + " WHERE id=" + be_id + ";")) {
            if (!rs.next()) {
                throw new Exception("Zur Berechnung NR." + be_id + " gibt es keinen Datenbankeintrag!");
            }

            int status = rs.getInt(2);
            long millis = rs.getLong(3);
            int max_t = rs.getInt(4);
            int max_s = rs.getInt(5);

            sc_sol = new D_Schule_SOL(2, max_t, max_s, status, millis);
        }

        // berechnung_lehrkraft (be_id, le_id, kuerzel, anzeigename)
        try (ResultSet rs = executeQuery("SELECT be_id, le_id, kuerzel, anzeigename, darf_vertreten FROM berechnung_lehrkraft WHERE be_id=" + be_id + ";")) {
            while (rs.next()) {
                int id = rs.getInt(2);
                D_Lehrkraft_SOL lesol = new D_Lehrkraft_SOL(id, rs.getString(3), rs.getString(4), rs.getInt(5));
                sc_sol.myLe.put(id, lesol);
            }
        }

        // berechnung_klasse (be_id, kl_id, kuerzel)
        try (ResultSet rs = executeQuery("SELECT be_id, kl_id, kuerzel FROM berechnung_klasse WHERE be_id=" + be_id + ";")) {
            while (rs.next()) {
                int kl_id = rs.getInt(2);
                String kuerzel = rs.getString(3);
                D_Klasse_SOL kl_sol = new D_Klasse_SOL(kl_id, kuerzel);
                sc_sol.myKl.put(kl_id, kl_sol);
            }
        }

        // berechnung_fach (be_id, fa_id, kuerzel, anzeigename)
        try (ResultSet rs = executeQuery("SELECT be_id, fa_id, kuerzel, anzeigename FROM berechnung_fach WHERE be_id=" + be_id + ";")) {
            while (rs.next()) {
                int fa_id = rs.getInt(2);
                String kuerzel = rs.getString(3);
                String anzeigename = rs.getString(4);
                D_Fach_SOL fa_sol = new D_Fach_SOL(fa_id, kuerzel, anzeigename);
                sc_sol.myFa.put(fa_id, fa_sol);
            }
        }

        // berechnung_raum (be_id, ra_id, kuerzel)
        try (ResultSet rs = executeQuery("SELECT be_id, ra_id, kuerzel FROM berechnung_raum WHERE be_id=" + be_id + ";")) {
            while (rs.next()) {
                int ra_id = rs.getInt(2);
                String kuerzel = rs.getString(3);
                D_Raum_SOL ra_sol = new D_Raum_SOL(ra_id, kuerzel);
                sc_sol.myRa.put(ra_id, ra_sol);
            }
        }

        // berechnung_kopplung (be_id, ko_id, kuerzel, anzeigename)
        try (ResultSet rs = executeQuery("SELECT be_id, ko_id, kuerzel, anzeigename FROM berechnung_kopplung WHERE be_id=" + be_id + ";")) {
            while (rs.next()) {
                int ko_id = rs.getInt(2);
                String kuerzel = rs.getString(3);
                String anzeigename = rs.getString(4);
                D_Kopplung_SOL ko_sol = new D_Kopplung_SOL(ko_id, kuerzel, anzeigename);
                sc_sol.myKo.put(ko_id, ko_sol);
            }
        }

        // berechnung_kurs (be_id, ko_id, ergaenzung)
        try (ResultSet rs = executeQuery("SELECT be_id, ku_id, ergaenzung, vertretungsreserven FROM berechnung_kurs WHERE be_id=" + be_id + ";")) {
            while (rs.next()) {
                int ku_id = rs.getInt(2);
                String ergaenzung = rs.getString(3);
                int vertretungsreserven = rs.getInt(4);
                D_Kurs_SOL ku_er = new D_Kurs_SOL(ku_id, ergaenzung, vertretungsreserven);
                sc_sol.myKu.put(ku_id, ku_er);
            }
        }

        // berechnung_stunde (be_id, ko_id, w, t, s)
        try (ResultSet rs = executeQuery("SELECT be_id, ku_id, w, t, s FROM berechnung_stunde WHERE be_id=" + be_id + ";")) {
            while (rs.next()) {
                int ku_id = rs.getInt(2);
                int w = rs.getInt(3);
                int t = rs.getInt(4);
                int s = rs.getInt(5);
                D_Stunde_SOL st_sol = new D_Stunde_SOL(w, t, s);
                sc_sol.myKu.get_obj_with_id(ku_id).mySt.addLast(st_sol);
            }
        }

        // berechnung_kurs_lehrkraft (be_id, ko_id, le_id)
        try (ResultSet rs = executeQuery("SELECT be_id, ku_id, le_id FROM berechnung_kurs_lehrkraft WHERE be_id=" + be_id + ";")) {
            while (rs.next()) {
                int ku_id = rs.getInt(2);
                int le_id = rs.getInt(3);
                D_Kurs_SOL ku_sol = sc_sol.myKu.get_obj_with_id(ku_id);
                D_Lehrkraft_SOL le_sol = sc_sol.myLe.get_obj_with_id(le_id);
                le_sol.myKu.addLast(ku_sol);
                ku_sol.myLe.addLast(le_sol);
            }
        }

        // berechnung_kurs_hosp (be_id, ko_id, le_id)
        try (ResultSet rs = executeQuery("SELECT be_id, ku_id, le_id FROM berechnung_kurs_hosp WHERE be_id=" + be_id + ";")) {
            while (rs.next()) {
                int ku_id = rs.getInt(2);
                int le_id = rs.getInt(3);
                D_Kurs_SOL ku_sol = sc_sol.myKu.get_obj_with_id(ku_id);
                D_Lehrkraft_SOL le_sol = sc_sol.myLe.get_obj_with_id(le_id);
                ku_sol.myLeHosp.addLast(le_sol);
                le_sol.myKuHosp.addLast(ku_sol);
            }
        }

        // berechnung_kurs_klasse (be_id, ko_id, kl_id)
        try (ResultSet rs = executeQuery("SELECT be_id, ku_id, kl_id FROM berechnung_kurs_klasse WHERE be_id=" + be_id + ";")) {
            while (rs.next()) {
                int ku_id = rs.getInt(2);
                int kl_id = rs.getInt(3);
                D_Kurs_SOL ku_sol = sc_sol.myKu.get_obj_with_id(ku_id);
                D_Klasse_SOL kl_sol = sc_sol.myKl.get_obj_with_id(kl_id);
                kl_sol.myKu.addLast(ku_sol);
                ku_sol.myKl.addLast(kl_sol);
            }
        }

        // berechnung_kurs_fach (be_id, ko_id, fa_id)
        try (ResultSet rs = executeQuery("SELECT be_id, ku_id, fa_id FROM berechnung_kurs_fach WHERE be_id=" + be_id + ";")) {
            while (rs.next()) {
                int ku_id = rs.getInt(2);
                int fa_id = rs.getInt(3);
                D_Kurs_SOL ku_sol = sc_sol.myKu.get_obj_with_id(ku_id);
                D_Fach_SOL fa_sol = sc_sol.myFa.get_obj_with_id(fa_id);
                fa_sol.myKu.addLast(ku_sol);
                ku_sol.myFa.addLast(fa_sol);
            }
        }

        // berechnung_kurs_raum (be_id, ko_id, ra_id)
        try (ResultSet rs = executeQuery("SELECT be_id, ku_id, ra_id FROM berechnung_kurs_raum WHERE be_id=" + be_id + ";")) {
            while (rs.next()) {
                int ku_id = rs.getInt(2);
                int ra_id = rs.getInt(3);
                D_Kurs_SOL ku_sol = sc_sol.myKu.get_obj_with_id(ku_id);
                D_Raum_SOL ra_sol = sc_sol.myRa.get_obj_with_id(ra_id);
                ra_sol.myKu.addLast(ku_sol);
                ku_sol.myRa.addLast(ra_sol);
            }
        }

        // berechnung_kurs_kopplung (be_id, ko_id, ko_id)
        try (ResultSet rs = executeQuery("SELECT be_id, ku_id, ko_id FROM berechnung_kurs_kopplung WHERE be_id=" + be_id + ";")) {
            while (rs.next()) {
                int ku_id = rs.getInt(2);
                int ko_id = rs.getInt(3);
                D_Kurs_SOL ku_sol = sc_sol.myKu.get_obj_with_id(ku_id);
                D_Kopplung_SOL ko_sol = sc_sol.myKo.get_obj_with_id(ko_id);
                ko_sol.myKu.addLast(ku_sol);
                ku_sol.myKo.addLast(ko_sol);
            }
        }

        return sc_sol;
    }

    void db_be_beenden(int id) throws Exception {
        if (map_process.containsKey(id)) {
            D_Worker worker = map_process.get(id);
            worker.beenden();
        }
    }

    void db_be_alle_beenden_und_loeschen() throws Exception {
        // Alle Berechnungen beenden...
        for (int row = 0; row < map_be.size(); row++) {
            D_Berechnung be = map_be.get_obj_with_row(row);
            db_be_beenden(be.id);
        }
        // Löschen...
        db_executeUpdate("DELETE FROM berechnung_schule;");
        db_sc_reload(false);
    }

    void db_be_loesche(D_Berechnung be) throws Exception {
        // Beenden...
        db_be_beenden(be.id);
        // Löschen...
        db_executeUpdate("DELETE FROM berechnung_schule WHERE id=" + be.id);
        db_sc_reload(false);
    }

    // ########## DB PL ##########
    void db_pl_save(String plan_table, String plan_id_column, int plan_id, String plan_data_column, String plan_data) throws Exception {
        db_executeUpdate("UPDATE " + plan_table + " SET " + plan_data_column + "='" + plan_data + "' WHERE " + plan_id_column + "=" + plan_id + ";");
        db_sc_reload(false);
    }

    // ### ANDERES ###
    void handle_exception(Exception ex) {
        gui.handle_exception(ex);
    }

    TreePath tr_get_node_einstellungen() {
        return new TreePath(new Object[]{this, d_einstellungen});
    }

    TreePath tr_get_node_root() {
        return new TreePath(new Object[]{this});
    }

    TreePath tr_get_node_schule() {
        return new TreePath(new Object[]{this, d_schule});
    }

    TreePath tr_get_node_lehrkraefte() {
        return new TreePath(new Object[]{this, d_schule, d_lehrkraefte});
    }

    TreePath tr_get_node_klassen() {
        return new TreePath(new Object[]{this, d_schule, d_klassen});
    }

    TreePath tr_get_node_faecher() {
        return new TreePath(new Object[]{this, d_schule, d_faecher});
    }

    TreePath tr_get_node_raeume() {
        return new TreePath(new Object[]{this, d_schule, d_raeume});
    }

    TreePath tr_get_node_kopplungen() {
        return new TreePath(new Object[]{this, d_schule, d_kopplungen});
    }

    TreePath tr_get_node_kurse() {
        return new TreePath(new Object[]{this, d_schule, d_kurse});
    }

    TreePath tr_get_node_lehrkraft(int id) {
        D_Lehrkraft le = map_le.get_obj_with_id(id);
        return new TreePath(new Object[]{this, d_schule, d_lehrkraefte, le});
    }

    TreePath tr_get_node_klasse(int id) {
        D_Klasse kl = map_kl.get_obj_with_id(id);
        return new TreePath(new Object[]{this, d_schule, d_klassen, kl});
    }

    TreePath tr_get_node_fach(int id) {
        D_Fach fa = map_fa.get_obj_with_id(id);
        return new TreePath(new Object[]{this, d_schule, d_faecher, fa});
    }

    TreePath tr_get_node_raum(int id) {
        D_Raum ra = map_ra.get_obj_with_id(id);
        return new TreePath(new Object[]{this, d_schule, d_raeume, ra});
    }

    TreePath tr_get_node_kopplung(int id) {
        D_Kopplung ko = map_ko.get_obj_with_id(id);
        return new TreePath(new Object[]{this, d_schule, d_kopplungen, ko});
    }

    TreePath tr_get_node_kurs(int id) {
        D_Kurs ku = map_ku.get_obj_with_id(id);
        return new TreePath(new Object[]{this, d_schule, d_kurse, ku});
    }

    TreePath tr_get_node_stunde(int id) {
        D_Stunde st = map_st.get_obj_with_id(id);
        D_Kurs ku = map_ku.get_obj_with_id(st.ku_id);
        return new TreePath(new Object[]{this, d_schule, d_kurse, ku, st});
    }

    TreePath tr_get_node_stunde_ko(int id) {
        D_Stunde_Ko st_ko = map_st_ko.get_obj_with_id(id);
        D_Kopplung ko = map_ko.get_obj_with_id(st_ko.ko_id);
        return new TreePath(new Object[]{this, d_schule, d_kopplungen, ko, st_ko});
    }

    TreePath tr_get_node_berechnung(int id) {
        D_Berechnung be = map_be.get_obj_with_id(id);
        return new TreePath(new Object[]{this, d_schule, d_berechnungen, be});
    }

    TreePath tr_get_node_berechnungen() {
        return new TreePath(new Object[]{this, d_schule, d_berechnungen});
    }

    boolean click_test_solver(JTextArea ta, int cpus) {

        for (int n = 1; n <= 128; n *= 2) {
            long m1 = System.currentTimeMillis();
            boolean b = test_solver_exact_1(n, cpus);
            long m2 = System.currentTimeMillis();
            ta.append("test_solver_exact_1(" + n + ") = " + b + " (" + (m2 - m1) + " Millisekunden)\n");
            if (!b) {
                return false;
            }
        }
        ta.append("\n");

        for (int n = 1; n <= 128; n *= 2) {
            long m1 = System.currentTimeMillis();
            boolean b = test_solver_pigeon_sat(n, cpus);
            long m2 = System.currentTimeMillis();
            ta.append("test_solver_pigeon_sat(" + n + ") = " + b + " (" + (m2 - m1) + " Millisekunden)\n");
            if (!b) {
                return false;
            }
        }
        ta.append("\n");

        for (int n = 1; n <= 11; n++) {
            long m1 = System.currentTimeMillis();
            boolean b = test_solver_pigeon_unsat(n, cpus);
            long m2 = System.currentTimeMillis();
            ta.append("test_solver_pigeon_unsat(" + n + ") = " + b + " (" + (m2 - m1) + " Millisekunden)\n");
            if (!b) {
                return false;
            }
        }
        ta.append("\n");

        return true;
    }

    private boolean test_solver_exact_1(int n, int cpus) {
        try {
            String pfad = Config.get_solver_crypto_pfad();

            SatWrapper sw1 = new SatWrapper(null, cpus, 0, 0);
            @SuppressWarnings("MismatchedReadAndWriteOfArray")
            int[] array1 = sw1.c_new_vars_array_exact(n, 1);

            SatWrapper sw2 = new SatWrapper(pfad, cpus, sw1.get_vars(), sw1.get_clauses());
            int[] array2 = sw2.c_new_vars_array_exact(n, 1);
            sw2.solve();

            int sum = 0;
            for (int i = 0; i < array2.length; i++) {
                int var = array2[i];
                if (sw2.is_true(var)) {
                    sum++;
                }
            }
            if (sum != 1) {
                return false;
            }

            return true;
        } catch (Exception ex) {
            gui.handle_exception(ex);
            return false;
        }
    }

    private boolean test_solver_pigeon_sat(int n, int cpus) {

        try {
            String pfad = Config.get_solver_crypto_pfad();

            SatWrapper sw1 = new SatWrapper(null, cpus, 0, 0);
            sw1.c_create_pigeon(n, n);

            SatWrapper sw2 = new SatWrapper(pfad, cpus, sw1.get_vars(), sw1.get_clauses());
            sw2.c_create_pigeon(n, n);

            return sw2.solve() >= 1;
        } catch (Exception ex) {
            gui.handle_exception(ex);
            return false;
        }
    }

    private boolean test_solver_pigeon_unsat(int n, int cpus) {

        try {
            String pfad = Config.get_solver_crypto_pfad();

            SatWrapper sw1 = new SatWrapper(null, cpus, 0, 0);
            sw1.c_create_pigeon(n + 1, n);

            SatWrapper sw2 = new SatWrapper(pfad, cpus, sw1.get_vars(), sw1.get_clauses());
            sw2.c_create_pigeon(n + 1, n);

            return sw2.solve() <= -1;
        } catch (Exception ex) {
            gui.handle_exception(ex);
            return false;
        }
    }

    D_Schule get_d_schule() {
        return d_schule;
    }

    boolean tr_path_exists(TreePath tp) {
        try {
            Object[] path = tp.getPath();
            for (int i = 1; i < path.length; i++) {
                Object parent = path[i - 1];
                Object child = path[i];
                if (!this.getHasChild(parent, child)) {
                    return false;
                }
            }
            return true;
        } catch (Exception ex) {
            handle_exception(ex);
        }
        return false;
    }

    final class TableModelKursLehrer extends DefaultTableModel {

        private D_Kurs ku = null;

        public TableModelKursLehrer() {
            this.addColumn("Lehrkraft");
            this.addColumn("✅");
        }

        @Override
        public Class getColumnClass(int columnIndex) {
            if (columnIndex == 1) {
                return Boolean.class;
            }
            if (columnIndex == 2) {
                return Integer.class;
            }
            return String.class;
        }

        @Override
        public int getColumnCount() {
            return 3;
        }

        @Override
        public boolean isCellEditable(int row, int column) {
            return column == 1;
        }

        @Override
        public void setValueAt(Object aValue, int row, int column) {
            // Zuerst Wert setzen...
            super.setValueAt(aValue, row, column);

            try {
                boolean checked = (Boolean) getValueAt(row, 1);
                int ku_id = ku.id;
                int le_id = (Integer) getValueAt(row, 2);
                if (checked) {
                    db_executeUpdate("INSERT INTO kurs_lehrkraft VALUES(" + ku_id + ", " + le_id + ");");
                } else {
                    db_executeUpdate("DELETE FROM kurs_lehrkraft WHERE kurs_id=" + ku_id + " AND lehrkraft_id=" + le_id + ";");
                }
                db_sc_reload(false);
                gui.gui_select_path(tr_get_node_kurs(ku.id), false, false);
            } catch (Exception ex) {
                handle_exception(ex);
            }

        }

        void lade_tabelle(D_Kurs ku) throws Exception {
            this.ku = ku;

            setRowCount(0); // GUI Tabelle leeren

            for (D_Lehrkraft le : map_le.rows()) {
                if (le.mapKu2.containsKey(ku.id)) {
                    addRow(new Object[]{le.kuerzel, true, le.id}); // GUI Tabelle füllen
                }
            }

            for (D_Lehrkraft le : map_le.rows()) {
                if (!le.mapKu2.containsKey(ku.id)) {
                    addRow(new Object[]{le.kuerzel, false, le.id}); // GUI Tabelle füllen
                }
            }
        }

    }

    final class TableModelKursKlasse extends DefaultTableModel {

        private D_Kurs ku = null;

        public TableModelKursKlasse() {
            this.addColumn("Klasse");
            this.addColumn("✅");
        }

        @Override
        public Class getColumnClass(int columnIndex) {
            if (columnIndex == 1) {
                return Boolean.class;
            }
            if (columnIndex == 2) {
                return Integer.class;
            }
            return String.class;
        }

        @Override
        public int getColumnCount() {
            return 3;
        }

        @Override
        public boolean isCellEditable(int row, int column) {
            return column == 1;
        }

        @Override
        public void setValueAt(Object aValue, int row, int column) {
            // Zuerst Wert setzen...
            super.setValueAt(aValue, row, column);

            try {
                boolean checked = (Boolean) getValueAt(row, 1);
                int ku_id = ku.id;
                int kl_id = (Integer) getValueAt(row, 2);
                if (checked) {
                    db_executeUpdate("INSERT INTO kurs_klasse VALUES(" + ku_id + ", " + kl_id + ");");
                } else {
                    db_executeUpdate("DELETE FROM kurs_klasse WHERE kurs_id=" + ku_id + " AND klasse_id=" + kl_id + ";");
                }
                db_sc_reload(false);
                gui.gui_select_path(tr_get_node_kurs(ku.id), false, false);
            } catch (Exception ex) {
                handle_exception(ex);
            }

        }

        void lade_tabelle(D_Kurs p_ku) throws Exception {
            this.ku = p_ku;

            setRowCount(0); // GUI Tabelle leeren

            for (D_Klasse kl : map_kl.rows()) {
                if (kl.mapKu.containsKey(ku.id)) {
                    addRow(new Object[]{kl.kuerzel, true, kl.id}); // GUI Tabelle füllen
                }
            }

            for (D_Klasse kl : map_kl.rows()) {
                if (!kl.mapKu.containsKey(ku.id)) {
                    addRow(new Object[]{kl.kuerzel, false, kl.id}); // GUI Tabelle füllen
                }
            }

        }

    }

    final class TableModelKursFach extends DefaultTableModel {

        private D_Kurs ku = null;

        public TableModelKursFach() {
            this.addColumn("Fach");
            this.addColumn("✅");
        }

        @Override
        public Class getColumnClass(int columnIndex) {
            if (columnIndex == 1) {
                return Boolean.class;
            }
            if (columnIndex == 2) {
                return Integer.class;
            }
            return String.class;
        }

        @Override
        public int getColumnCount() {
            return 3;
        }

        @Override
        public boolean isCellEditable(int row, int column) {
            return column == 1;
        }

        @Override
        public void setValueAt(Object aValue, int row, int column) {
            // Zuerst Wert setzen...
            super.setValueAt(aValue, row, column);

            try {
                boolean checked = (Boolean) getValueAt(row, 1);
                int ku_id = ku.id;
                int fa_id = (Integer) getValueAt(row, 2);
                if (checked) {
                    db_executeUpdate("DELETE FROM kurs_fach WHERE kurs_id=" + ku_id + ";");
                    db_executeUpdate("INSERT INTO kurs_fach VALUES(" + ku_id + ", " + fa_id + ");");
                } else {
                    db_executeUpdate("DELETE FROM kurs_fach WHERE kurs_id=" + ku_id + " AND fach_id=" + fa_id + ";");
                }
                db_sc_reload(false);
                gui.gui_select_path(tr_get_node_kurs(ku.id), false, false);
            } catch (Exception ex) {
                handle_exception(ex);
            }

        }

        void lade_tabelle(D_Kurs p_ku) throws Exception {
            this.ku = p_ku;

            setRowCount(0); // GUI Tabelle leeren

            for (D_Fach fa : map_fa.rows()) {
                if (fa.mapKu.containsKey(ku.id)) {
                    addRow(new Object[]{fa.kuerzel, true, fa.id}); // GUI Tabelle füllen
                }
            }

            for (D_Fach fa : map_fa.rows()) {
                if (!fa.mapKu.containsKey(ku.id)) {
                    addRow(new Object[]{fa.kuerzel, false, fa.id}); // GUI Tabelle füllen
                }
            }

        }

    }

    final class TableModelKursRaum extends DefaultTableModel {

        private D_Kurs ku = null;

        public TableModelKursRaum() {
            this.addColumn("Raum");
            this.addColumn("✅");
        }

        @Override
        public Class getColumnClass(int columnIndex) {
            if (columnIndex == 1) {
                return Boolean.class;
            }
            if (columnIndex == 2) {
                return Integer.class;
            }
            return String.class;
        }

        @Override
        public int getColumnCount() {
            return 3;
        }

        @Override
        public boolean isCellEditable(int row, int column) {
            return column == 1;
        }

        @Override
        public void setValueAt(Object aValue, int row, int column) {
            // Zuerst Wert setzen...
            super.setValueAt(aValue, row, column);

            try {
                boolean checked = (Boolean) getValueAt(row, 1);
                int ku_id = ku.id;
                int ra_id = (Integer) getValueAt(row, 2);
                if (checked) {
                    db_executeUpdate("INSERT INTO kurs_raum VALUES(" + ku_id + ", " + ra_id + ");");
                } else {
                    db_executeUpdate("DELETE FROM kurs_raum WHERE kurs_id=" + ku_id + " AND raum_id=" + ra_id + ";");
                }

                db_sc_reload(false);
                gui.gui_select_path(tr_get_node_kurs(ku.id), false, false);
            } catch (Exception ex) {
                handle_exception(ex);
            }

        }

        void lade_tabelle(D_Kurs p_ku) throws Exception {
            this.ku = p_ku;

            setRowCount(0); // GUI Tabelle leeren

            for (D_Raum ra : map_ra.rows()) {
                if (ra.mapKu.containsKey(ku.id)) {
                    addRow(new Object[]{ra.kuerzel, true, ra.id}); // GUI Tabelle füllen
                }
            }

            for (D_Raum ra : map_ra.rows()) {
                if (!ra.mapKu.containsKey(ku.id)) {
                    addRow(new Object[]{ra.kuerzel, false, ra.id}); // GUI Tabelle füllen
                }
            }

        }

    }

    final class TableModelKursKopplung extends DefaultTableModel {

        private D_Kurs ku = null;

        public TableModelKursKopplung() {
            this.addColumn("Kopplung");
            this.addColumn("✅");
        }

        @Override
        public Class getColumnClass(int columnIndex) {
            if (columnIndex == 1) {
                return Boolean.class;
            }
            if (columnIndex == 2) {
                return Integer.class;
            }
            return String.class;
        }

        @Override
        public int getColumnCount() {
            return 3;
        }

        @Override
        public boolean isCellEditable(int row, int column) {
            return column == 1;
        }

        @Override
        public void setValueAt(Object aValue, int row, int column) {
            // Zuerst Wert setzen...
            super.setValueAt(aValue, row, column);

            try {
                boolean checked = (Boolean) getValueAt(row, 1);
                int ku_id = ku.id;
                int ko_id = (Integer) getValueAt(row, 2);
                if (checked) {
                    db_executeUpdate("DELETE FROM kurs_kopplung WHERE kurs_id=" + ku_id + ";");
                    db_executeUpdate("INSERT INTO kurs_kopplung VALUES(" + ku_id + ", " + ko_id + ");");
                } else {
                    db_executeUpdate("DELETE FROM kurs_kopplung WHERE kurs_id=" + ku_id + " AND kopplung_id=" + ko_id + ";");
                }

                db_sc_reload(false);
                gui.gui_select_path(tr_get_node_kurs(ku.id), false, false);
            } catch (Exception ex) {
                handle_exception(ex);
            }

        }

        void lade_tabelle(D_Kurs p_ku) throws Exception {
            this.ku = p_ku;

            setRowCount(0); // GUI Tabelle leeren

            for (D_Kopplung ko : map_ko.rows()) {
                if (ko.mapKu.containsKey(ku.id)) {
                    addRow(new Object[]{ko.kuerzel, true, ko.id}); // GUI Tabelle füllen
                }
            }

            for (D_Kopplung ko : map_ko.rows()) {
                if (!ko.mapKu.containsKey(ku.id)) {
                    addRow(new Object[]{ko.kuerzel, false, ko.id}); // GUI Tabelle füllen
                }
            }

        }

    }

    final class TableModelKursKurs extends DefaultTableModel {

        private D_Kurs ku = null;

        public TableModelKursKurs() {
            this.addColumn("Nicht gleicher Tag mit");
            this.addColumn("✅");
        }

        @Override
        public Class getColumnClass(int columnIndex) {
            if (columnIndex == 1) {
                return Boolean.class;
            }
            if (columnIndex == 2) {
                return Integer.class;
            }
            return String.class;
        }

        @Override
        public int getColumnCount() {
            return 3;
        }

        @Override
        public boolean isCellEditable(int row, int column) {
            return column == 1;
        }

        @Override
        public void setValueAt(Object aValue, int row, int column) {
            // Zuerst Wert setzen...
            super.setValueAt(aValue, row, column);

            try {
                boolean checked = (Boolean) getValueAt(row, 1);
                int ku_id_1 = ku.id;
                int ku_id_2 = (Integer) getValueAt(row, 2);
                if (checked) {
                    db_executeUpdate("INSERT INTO kurs_kurs VALUES(" + ku_id_1 + ", " + ku_id_2 + ");");
                    db_executeUpdate("INSERT INTO kurs_kurs VALUES(" + ku_id_2 + ", " + ku_id_1 + ");");
                } else {
                    db_executeUpdate("DELETE FROM kurs_kurs WHERE kurs_id_1=" + ku_id_1 + " AND kurs_id_2=" + ku_id_2 + ";");
                    db_executeUpdate("DELETE FROM kurs_kurs WHERE kurs_id_1=" + ku_id_2 + " AND kurs_id_2=" + ku_id_1 + ";");
                }
                db_sc_reload(false);
                gui.gui_select_path(tr_get_node_kurs(ku.id), false, false);
            } catch (Exception ex) {
                handle_exception(ex);
            }

        }

        void lade_tabelle(D_Kurs ku) throws Exception {
            this.ku = ku;

            setRowCount(0); // GUI Tabelle leeren

            for (D_Kurs ku2 : map_ku.rows()) {
                if (ku2.id != ku.id) {
                    if (ku2.nicht_gleicher_tag_wie.containsKey(ku.id)) {
                        addRow(new Object[]{ku2.toString(), true, ku2.id}); // GUI Tabelle füllen
                    }
                }
            }

            for (D_Kurs ku2 : map_ku.rows()) {
                if (ku2.id != ku.id) {
                    if (!ku2.nicht_gleicher_tag_wie.containsKey(ku.id)) {
                        addRow(new Object[]{ku2.toString(), false, ku2.id}); // GUI Tabelle füllen
                    }
                }
            }

        }

    }

    final class TableModelKursHosp extends DefaultTableModel {

        private D_Kurs ku = null;

        public TableModelKursHosp() {
            this.addColumn("Hospitiert von");
            this.addColumn("✅");
        }

        @Override
        public Class getColumnClass(int columnIndex) {
            if (columnIndex == 1) {
                return Boolean.class;
            }
            if (columnIndex == 2) {
                return Integer.class;
            }
            return String.class;
        }

        @Override
        public int getColumnCount() {
            return 3;
        }

        @Override
        public boolean isCellEditable(int row, int column) {
            return column == 1;
        }

        @Override
        public void setValueAt(Object aValue, int row, int column) {
            // Zuerst Wert setzen...
            super.setValueAt(aValue, row, column);

            try {
                boolean checked = (Boolean) getValueAt(row, 1);
                int ku_id = ku.id;
                int le_id = (Integer) getValueAt(row, 2);
                if (checked) {
                    db_executeUpdate("INSERT INTO kurs_hosp VALUES(" + ku_id + ", " + le_id + ");");
                } else {
                    db_executeUpdate("DELETE FROM kurs_hosp WHERE kurs_id=" + ku_id + " AND lehrkraft_id=" + le_id + ";");
                }
                db_sc_reload(false);
                gui.gui_select_path(tr_get_node_kurs(ku.id), false, false);
            } catch (Exception ex) {
                handle_exception(ex);
            }

        }

        void lade_tabelle(D_Kurs ku) throws Exception {
            this.ku = ku;

            setRowCount(0); // GUI Tabelle leeren

            for (D_Lehrkraft le : map_le.rows()) {
                if (le.mapKuHosp2.containsKey(ku.id)) {
                    addRow(new Object[]{le.kuerzel, true, le.id}); // GUI Tabelle füllen
                }
            }

            for (D_Lehrkraft le : map_le.rows()) {
                if (!le.mapKuHosp2.containsKey(ku.id)) {
                    addRow(new Object[]{le.kuerzel, false, le.id}); // GUI Tabelle füllen
                }
            }
        }

    }

    private final class D_Worker extends Thread {
        //final SatWrapper sw;

        final int be_id;
        final D_Schule_SAT sc_sat;
        final String pfad;
        final int cpus = 1;
        Process proc = null;

        public D_Worker(int p_be_id, D_Schule_SAT p_sc_sat, String p_pfad) {
            be_id = p_be_id;
            sc_sat = p_sc_sat;
            pfad = p_pfad;
        }

        @Override
        public void run() {
            try {
                // Simuliere die Formel, um die Variablen und Klauseln zu zählen.
                long t1 = System.currentTimeMillis();
                SatWrapper sw1 = new SatWrapper(null, cpus, 0, 0);
                cnf(sw1);

                // Erzeuge die Formel erneut.
                SatWrapper sw2 = new SatWrapper(pfad, cpus, sw1.get_vars(), sw1.get_clauses());
                proc = sw2.get_process();
                synchronized (this) {
                    notify();
                }
                cnf(sw2);

                // Starte die Berechnung
                int be_status = sw2.solve();
                long t6 = System.currentTimeMillis();

                long be_millis = t6 - t1;
                SwingUtilities.invokeLater(new Runnable() {
                    @Override
                    public void run() {
                        try {
                            db_be_fertig(be_id, be_status, be_millis, sw2, sc_sat);
                        } catch (Exception ex) {
                            handle_exception(ex);
                        }
                    }
                });

            } catch (Exception ex) {
                handle_exception(ex);
            }
        }

        private synchronized void cnf(SatWrapper sw) throws Exception {
            sc_sat.cnf_pre(sw);
            sc_sat.cnf_post(sw);
        }

        private synchronized void beenden() throws InterruptedException {
            if (proc == null) {
                wait();
            }
            proc.destroy();
        }
    }

}
