package bar.stundenplan;

import java.util.HashMap;
import java.util.LinkedList;

/**
 * @author Benjamin A. Bartsch
 * @param <T>
 */
public class RowMap<T> {

    private final HashMap<Integer, T> map_id_to_obj;
    private final HashMap<Integer, T> map_row_to_obj;
    private final HashMap<T, Integer> map_obj_to_row;

    public RowMap() {
        map_id_to_obj = new HashMap<>();
        map_row_to_obj = new HashMap<>();
        map_obj_to_row = new HashMap<>();
    }

    public void clear() {
        map_id_to_obj.clear();
        map_row_to_obj.clear();
        map_obj_to_row.clear();
    }

    public boolean isEmpty() {
        return map_id_to_obj.isEmpty();
    }

    public T get_obj_with_id(int id) {
        return map_id_to_obj.get(id);
    }

    public T get_obj_with_row(int row) {
        return map_row_to_obj.get(row);
    }

    public int get_row_of_obj(T obj) {
        return map_obj_to_row.get(obj);
    }

    int size() {
        return map_id_to_obj.size();
    }

    void put(int id, T obj) {
        int row = map_id_to_obj.size();
        map_id_to_obj.put(id, obj);
        set_row_of_obj(row, obj);
    }

    void remove(int id) {
        T obj = map_id_to_obj.get(id);
        int row = map_obj_to_row.get(obj);
        map_id_to_obj.remove(id);
        map_obj_to_row.remove(obj);
        map_row_to_obj.remove(row);
    }

    void set_row_of_obj(int row, T obj) {
        map_row_to_obj.put(row, obj);
        map_obj_to_row.put(obj, row);
    }

    LinkedList<T> rows() {
        LinkedList<T> list = new LinkedList<>();
        for (int row = 0; row < map_row_to_obj.size(); row++) {
            list.addLast(map_row_to_obj.get(row));
        }
        return list;
    }

    boolean contains(T obj) {
        return map_obj_to_row.containsKey(obj);
    }

}
